import numpy as np

from artemis.ml.datasets.datasets import DataSet, DataCollection


__author__ = 'peter'


sigm = lambda x: 1./(1+np.exp(-x))


def get_logistic_regression_dataset(**kwargs):
    x_tr, y_tr, x_ts, y_ts, _ = get_logistic_regression_data(**kwargs)
    return LogisticRegressionDataset(**kwargs)


class LogisticRegressionDataset(DataSet):

    def __init__(self, **kwargs):
        x_tr, y_tr, x_ts, y_ts, w_true = get_logistic_regression_data(**kwargs)
        DataSet.__init__(self, DataCollection(x_tr, y_tr), DataCollection(x_ts, y_ts))
        self._w_true = w_true

    @property
    def w_true(self):
        return self._w_true


def get_logistic_regression_data(
        n_dims = 20,
        n_training = 1000,
        n_test = 100,
        noise_factor = 1,
        seed = 5354355
        ):
    rng = np.random.RandomState(seed)
    n_samples = n_training+n_test
    # Since magnitude of x.dot(w) grows with square-root of len(x), we scale x depending on n_dims
    x_scale = np.array(1.)/noise_factor if noise_factor!=0 else float('inf')
    x = ((rng.rand(n_samples, n_dims) > 0.5)*2-1)
    w = rng.rand(n_dims, 1)
    z = sigm(x.dot(w)*x_scale)
    y = (z > rng.rand(*z.shape)).astype(int)
    return x[:n_training], y[:n_training], x[n_training:], y[n_training:], w



from collections import OrderedDict

import itertools
import numpy as np

__author__ = 'peter'


def flatten_struct(struct, primatives = (int, float, np.ndarray, basestring, bool), custom_handlers = {},
        break_into_objects = True, detect_duplicates = True, memo = None):
    """
    Given some nested struct, return a list<*(str, primative)>, where primative
    is some some kind of object that you don't break down any further, and str is a
    string representation of how you would access that propery from the root object.

    Don't try any fancy circular references here, it's not going to go well for you.

    :param struct: Something, anything.
    :param primatives: A list of classes that will not be broken into.
    :param custum_handlers: A dict<type:func> where func has the form data = func(obj).  These
        will be called if the type of the struct is in the dict of custom handlers.
    :param break_into_objects: True if you want to break into objects to see what's inside.
    :return: list<*(str, primative)>
    """
    if memo is None:
        memo = {}

    if id(struct) in memo:
        return [(None, memo[id(struct)])]
    elif detect_duplicates:
        memo[id(struct)] = 'Already Seen object at %s' % hex(id(struct))

    if isinstance(struct, primatives):
        return [(None, struct)]
    elif isinstance(struct, tuple(custom_handlers.keys())):
        handler = custom_handlers[custom_handlers.keys()[[isinstance(struct, t) for t in custom_handlers].index(True)]]
        return [(None, handler(struct))]
    elif isinstance(struct, dict):
        return sum([
            [("[%s]%s" % (("'%s'" % key if isinstance(key, str) else key), subkey if subkey is not None else ''), v)
                for subkey, v in flatten_struct(value, custom_handlers=custom_handlers, primatives=primatives, break_into_objects=break_into_objects, memo=memo, detect_duplicates=detect_duplicates)]
                for key, value in struct.iteritems()
            ], [])
    elif isinstance(struct, (list, tuple)):
        # for i, value in enumerate(struct):
        return sum([
            [("[%s]%s" % (i, subkey if subkey is not None else ''), v)
                for subkey, v in flatten_struct(value, custom_handlers=custom_handlers, primatives=primatives, break_into_objects=break_into_objects, memo=memo, detect_duplicates=detect_duplicates)]
                for i, value in enumerate(struct)
            ], [])
    elif struct is None or not hasattr(struct, '__dict__'):
        return []
    elif break_into_objects:  # It's some kind of object, lets break it down.
        return sum([
            [(".%s%s" % (key, subkey if subkey is not None else ''), v)
                for subkey, v in flatten_struct(value, custom_handlers=custom_handlers, primatives=primatives, break_into_objects=break_into_objects, memo=memo, detect_duplicates=detect_duplicates)]
                for key, value in struct.__dict__.iteritems()
            ], [])
    else:
        return [(None, memo[id(struct)])]


_primitive_containers = (list, tuple, dict, set)


def flatten_nested_object(data_object, containers = _primitive_containers, include_classes = False):
    """
    Given a data object, walk through the object and return a dict of contained objects.
    It's best to see the tests to understand this function.

    Don't try any fancy circular references here - it's not going to go well for you.

    :param data_object:
    :return: An OrderedDict containing, in depth-first-order, the list of objects and containers.
    """
    directory_listing = OrderedDict()
    if include_classes:
        directory_listing['__class__'] = type(data_object)
    if data_object in containers:
        if isinstance(data_object, (list, tuple)):
            directory_listing[None] = type(data_object)
            for i, x in enumerate(data_object):
                sub_obj = flatten_nested_object(data_object)
                for k, v in sub_obj.iteritems():
                    directory_listing[(i, )+k] = v
        elif isinstance(data_object, dict):
            for key, val in data_object.iteritems():
                sub_obj = flatten_nested_object(val)
                for k, v in sub_obj.iteritems():
                    directory_listing[(key, )+k] = v
        else:
            raise Exception('Unidentified container type: {}'.format(type(data_object)))
    else:
        directory_listing[()] = data_object


def get_meta_object(data_object, containers = _primitive_containers):
    """
    Given an arbitrary data structure, return a "meta object" which is the same structure, except all non-container
    objects are replaced by their types.

    e.g.
        get_meta_obj([1, 2, {'a':(3, 4), 'b':['hey', 'yeah']}, 'woo']) == [int, int, {'a':(int, int), 'b':[str, str]}, str]

    :param data_object: A data object with arbitrary nested structure
    :param containers: Which classes to consider containers.  Classes derived from these are also considered containers.
    :return:
    """
    if isinstance(data_object, containers):
        if isinstance(data_object, (list, tuple, set)):
            return type(data_object)(get_meta_object(x) for x in data_object)
        elif isinstance(data_object, dict):
            return type(data_object)((k, get_meta_object(v)) for k, v in data_object.iteritems())
    else:
        return type(data_object)


class NestedType(object):
    """
    An object which represents the type of an arbitrarily nested data structure.  It can be constructed directly
    from a nested type descriptor, or indirectly using the NestedType.from_data(...) constructor.

    For example
        NestedType.from_data([1, 2, {'a':(3, 4.), 'b':'c'}]) == NestedType([int, int, {'a':(int, float), 'b':str}])
    """

    def __init__(self, meta_object):
        """
        :param meta_object: A nested type descriptor.  See docstring and tests for examples.
        """
        self.meta_object = meta_object

    def is_type_for(self, data_object):
        return get_meta_object(data_object)==self.meta_object

    def check_type(self, data_object):
        """
        Assert that the data_object has a format matching this NestedType.  Throw a TypeError if it does not.
        :param data_object:
        :return:
        """
        if not self.is_type_for(data_object):
            raise TypeError('The data object has type {}, which does not match this format: {}'.format(NestedType.from_data(data_object), self))

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.meta_object)

    def __eq__(self, other):
        return self.meta_object == other.meta_object

    def get_leaves(self, data_object, check_types = True):
        """
        :param data_object: Given a nested object, get the "leaf" values in Depth-First Order
        :return: A list of leaf values.
        """
        if check_types:
            assert self.is_type_for(data_object)
        return get_leaf_values(data_object)

    def expand_from_leaves(self, leaves, check_types = True, assert_fully_used=True):
        return _fill_meta_object(self.meta_object, (x for x in leaves), check_types=check_types, assert_fully_used=assert_fully_used)

    @staticmethod
    def from_data(data_object, containers = _primitive_containers):
        """
        :param data_object: A nested data object
        :param containers: The list of classes defined as "containers"
        :return: A NestedType object
        """
        return NestedType(get_meta_object(data_object, containers=containers))


def get_leaf_values(data_object, containers = _primitive_containers):
    """
    Collect leaf values.
    Caution: If your data contains dicts, you may not get the same order of results when you call this function with
    different dict objects containing the same data.  Python only guarantees that a given dict will always iterate in
    the same order so long as it is not modified.  See https://docs.python.org/2/library/stdtypes.html#dict.items

    :param data_object: An arbitrary nested data object
    :param containers: The list of "container" classes that we should break into.
    :return:
    """
    leaf_values = []
    if isinstance(data_object, containers):
        if isinstance(data_object, (list, tuple)):
            leaf_values += [val for x in data_object for val in get_leaf_values(x)]
        elif isinstance(data_object, OrderedDict):
            leaf_values += [val for k, x in data_object.iteritems() for val in get_leaf_values(x)]
        elif isinstance(data_object, dict):
            leaf_values += [val for k in sorted(data_object.keys()) for val in get_leaf_values(data_object[k])]
        else:
            raise Exception('Have no way to consistently extract leaf values from a {}'.format(data_object))
        return leaf_values
    else:
        return [data_object]


def _fill_meta_object(meta_object, data_iteratable, assert_fully_used = True, check_types = True):
    """
    Fill the data from the iterable into the meta_object.
    :param meta_object: A nested type descripter.  See NestedType init
    :param data_iteratable: The iterable data object
    :param assert_fully_used: Assert that we actually get through all the items in the iterable
    :return: The filled object
    """

    try:
        if isinstance(meta_object, _primitive_containers):
            if isinstance(meta_object, (list, tuple, set)):
                filled_object = type(meta_object)(_fill_meta_object(x, data_iteratable, assert_fully_used=False, check_types=check_types) for x in meta_object)
            elif isinstance(meta_object, OrderedDict):
                filled_object = type(meta_object)((k, _fill_meta_object(val, data_iteratable, assert_fully_used=False, check_types=check_types)) for k, val in meta_object.iteritems())
            elif isinstance(meta_object, dict):
                filled_object = type(meta_object)((k, _fill_meta_object(meta_object[k], data_iteratable, assert_fully_used=False, check_types=check_types)) for k in sorted(meta_object.keys()))
            else:
                raise Exception('Cannot handle container type: "{}"'.format(type(meta_object)))
        else:
            next_data = data_iteratable.next()
            if check_types and meta_object is not type(next_data):
                raise TypeError('The type of the data object: {} did not match type from the meta object: {}'.format(type(next_data), meta_object))
            filled_object = next_data
    except StopIteration:
        raise TypeError('The data iterable you were going through ran out before the object {} could be filled.'.format(meta_object))

    if assert_fully_used:
        try:
            data_iteratable.next()
            raise TypeError('It appears that the data object you were using to fill your meta object had more data than could fit.')
        except StopIteration:
            pass
    return filled_object


def get_nested_value(data_object, key_chain):
    if len(key_chain)>0:
        return get_nested_value(data_object[key_chain[0]], key_chain=key_chain[1:])
    else:
        return data_object


class ExpandingDict(dict):

    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            self[key] = ExpandingDict()
            return dict.__getitem__(self, key)


class ExpandingOrderedDict(OrderedDict):

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            self[key] = ExpandingDict()
            return OrderedDict.__getitem__(self, key)


def expand_struct(struct):

    expanded_struct = ExpandingDict()

    for k in struct.keys():
        exec('expanded_struct%s = struct["%s"]' % (k, k))

    return expanded_struct


def seqstruct_to_structseq(seqstruct, as_arrays=False):
    """
    Turn a sequence of identically-structured nested objects into a nested object of sequences.
    :param seqstruct: A sequence (list or tuple) of nested objects with similar format
    :param as_arrays: Turn the output sequences into numpy arrays
    :return: A nested object with sequences

    For example, if you go:
        signal_seqs = seqstruct_to_structseq(seq_signals)
    Then
        frame_number = 5
        seq_signals[frame_number]['inputs']['camera'] == signal_seqs['inputs']['camera'][frame_number]
    """
    if len(seqstruct)==0:
        return []

    nested_type = NestedType.from_data(seqstruct[0])
    leaf_data = [nested_type.get_leaves(s) for s in seqstruct]
    batch_leaf_data = [np.array(d) for d in zip(*leaf_data)] if as_arrays else zip(*leaf_data)
    structseq = nested_type.expand_from_leaves(leaves = batch_leaf_data, check_types=False)
    return structseq


def structseq_to_seqstruct(structseq):
    """
    Turn a nested object of sequences into a sequence of identically-structured nested objects.

    This is the inverse of seqstruct_to_structseq

    :param structseq: A nested object with sequences
    :return: A sequence (list or tuple) of nested objects with similar format
    """
    nested_type = NestedType.from_data(structseq)
    leaf_data = nested_type.get_leaves(structseq, check_types=False)
    sequence = zip(*leaf_data)
    seqstruct = [nested_type.expand_from_leaves(s, check_types=False) for s in sequence]
    return seqstruct



from __future__ import print_function

import atexit
import os
import shlex
import signal
import subprocess
import sys
import threading
import time
import uuid
from ConfigParser import NoOptionError
import paramiko
from artemis.config import get_artemis_config_value
from artemis.fileman.config_files import get_config_value
from artemis.remote.utils import get_local_ips


class ParamikoPrintThread(threading.Thread):
        def __init__(self, source_pipe, target_pipe, stopping_criterium=None, stop_event=None, prefix=""):
            '''
            :param source_pipe: The ssh pipe from which to forward communications
            :param target_pipe: Either stdout or stderr. This determines if stderr or stdout is read from the ssh channel.
            :param stopping_criterium: function which takes in a line from source_pipe and evaluates to boolean.:
            :param prefix: A prefix that is attached to every printed line.
            :return:
            '''
            self.source_pipe = source_pipe
            self.prefix = prefix
            self.target_pipe = target_pipe
            self.stopping_criterium = stopping_criterium
            super(ParamikoPrintThread, self).__init__()

        def run(self, ):
            with self.source_pipe:
                # Terminates when source pipe runs dry.
                for line in iter(self.source_pipe.readline, b''):
                    self.target_pipe.write("%s%s"%(self.prefix,line))
                    self.target_pipe.flush()
                    if self.stopping_criterium is not None and self.stopping_criterium(line):
                        break


class Nanny(object):
    '''
    Manages child processes. This class manages the start, live and deconstruction of child processes across different machines.
    '''
    def __init__(self):
        self.child_processes = {}
        self.stdout_threads = {}
        self.original_sigint_handler = signal.getsignal(signal.SIGINT)
        self.original_sigterm_handler = signal.getsignal(signal.SIGTERM)
        signal.signal(signal.SIGINT, self.deconstruct)
        signal.signal(signal.SIGTERM, self.deconstruct)
        # atexit.register(self.deconstruct)

    def register_child_process(self, cp):
        '''
        This adds a child process to the Nanny.
        :return:
        '''
        self.child_processes[cp.get_id()] = cp

    def execute_all_child_processes(self, time_out=1, terminate_at_error=False,):
        '''
        Executes all child processes and starts managing communications. This method returns only when all child processes terminated.
        It might be the case that some child-processes hang or don't terminate. In this case, when one (the first) process
          terminates, all other processes are killed after time_out seconds. This behaviour is triggered also when

        :param: time_out
        :return:
        '''

        termination_request_event = threading.Event()
        stdout_stopping_criterium = lambda x: "Training finished" in x
        stdout_threads = {}
        stderr_threads = {}

        N_workers = 0
        def err_fun(cp_ip):
            print("Error in process on %s:"%cp_ip)
            return True
        name_max_lenght = max([len(cp.name) for cp in self.child_processes.values()])
        for id, cp in self.child_processes.iteritems():
            stdin, stdout, stderr = cp.execute_child_process()

            prefix = cp.name.ljust(name_max_lenght)+": "
            if "worker" in cp.name:
                N_workers+=1
            stdout_thread = threading.Thread(target=self.monitor_and_forward_child_communication,
                                  args=(stdout,sys.stdout,termination_request_event,stdout_stopping_criterium, prefix))

            stderr_stopping_criterium = lambda x: err_fun(cp.get_ip()) if terminate_at_error else None
            stderr_thread = threading.Thread(target=self.monitor_and_forward_child_communication,
                                  args=(stderr,sys.stderr,termination_request_event,stderr_stopping_criterium, prefix))
            stdout_threads[cp.get_id()] = stdout_thread
            stderr_threads[cp.get_id()] = stderr_thread

            stdout_thread.start()
            stderr_thread.start()

        try:
            for _ in range(N_workers):
                while not termination_request_event.wait(0.01):
                    pass
                termination_request_event.clear()
        except KeyboardInterrupt:
            sys.exit(1)

        # Grace perior for other threads to shutdown
        time.sleep(time_out)
        for id,cp in self.child_processes.iteritems():
            if cp.is_alive():
                print("Child Process %s at %s did not terminate %s seconds after the first process in cluster terminated. Terminating now." %(cp.get_name(), cp.get_ip(), time_out))
                cp.kill()

        for id,cp in self.child_processes.iteritems():
            if cp.is_alive():
                print("Child Process %s at %s did not terminate. Force quitting now." %(cp.get_name(),cp.get_ip()))
                cp.deconstruct(signal.SIGTERM)

        # This should return immediatly, since the unterlying pipes should have run dry. if it doesn't I messed up...:
        for stdout_thread, stderr_thread in zip(stdout_threads.values(), stderr_threads.values()):
            assert not stdout_thread.is_alive(), "This should not have happened"
            assert not stderr_thread.is_alive(), "This should not have happened"

    def deconstruct(self, signum, frame=None):
        '''
        This method is called when SIGINT or SIGTERM are called.
        This aggressively deconstructs the Nanny and all child processes. Then, the signal is passed back to the original signal handlers.
        :return:
        '''

        for cp in self.child_processes.values():
            cp.kill()
        time.sleep(1.0)
        for cp in self.child_processes.values():
            if cp.is_alive():
                print("Child Process %s at %s still alive, force terminating now"% (cp.name, cp.get_ip()))
                cp.kill(signal=signal.SIGTERM)

        signal.signal(signal.SIGINT, self.original_sigint_handler)
        signal.signal(signal.SIGTERM, self.original_sigterm_handler)
        os.kill(os.getpid(), signum)

    def monitor_and_forward_child_communication(self, source_pipe, target_pipe, termination_request_event=None, stopping_criterium=None,  prefix=""):
        '''
        thread to forward communication from source_pipe to target_pipe
        :param source_pipe:
        :param target_pipe:
        :param termination_request_event: Is set once the source_pipe has closed and this thread terminates, or when ths stopping_criterium has been met.
        :param stopping_criterium:
        :param prefix:
        :return:
        '''
        with source_pipe:
            for line in iter(source_pipe.readline, b''):
                target_pipe.write("%s%s"%(prefix,line))
                target_pipe.flush()
                if stopping_criterium is not None and stopping_criterium(line) and not "pydev debugger" in line and line.strip():
                    target_pipe.write(source_pipe.read())
                    target_pipe.flush()
                    break
            if termination_request_event is not None:
                termination_request_event.set() # The input pipe closed, this thread terminates and we would like everybody to terminate


class ChildProcess(object):
    counter=1
    def __init__(self, ip_address, command, name=None, take_care_of_deconstruct=False):
        '''
        Creates a ChildProcess
        :param ip_address: The command will be executed at this ip_address
        :param command: the command to execute. If it is a python command and remote, will substitute the virtualenv
        :param name: optional name. If not set, will be process_i, with i a clobal counter
        :param take_care_of_deconstruct: If set to True, deconstruct() is registered at exit
        :return:
        '''
        if name is None:
            name = "process_%s"%ChildProcess.counter
        ChildProcess.counter += 1
        self.name = name
        self.ip_address = ip_address
        self.local_process = self.ip_address in get_local_ips()


        # command = command.replace(unichr(34),unichr(39))
        self.command = command
        self.id = uuid.uuid4()
        self.channel = None
        self.cp_started = False
        self.take_care_of_deconstruct = take_care_of_deconstruct
        if self.take_care_of_deconstruct:
            atexit.register(self.deconstruct)

    def prepare_command(self,command):
        '''
        All the stuff that I need to prepare for the command to definitely work
        :param command:
        :return:
        '''
        if self.is_local():
            home_dir = os.path.expanduser("~")

        else:
            _,_,stdout,_ = self._run_command("echo $$; exec echo ~")
            home_dir = stdout.read().strip()

        if type(command) == list:
            if not self.local_process:
                command = [c.replace("python", self.get_extended_command(get_artemis_config_value(section=self.get_ip(), option="python")), 1) if c.startswith("python") else c for c in command]
                command = [s.replace("~",home_dir) for s in command]
                command = " ".join([c for c in command])
            else:
                command = [c.strip("'") for c in command]
                command = [c.replace("python", sys.executable, 1) if c.startswith("python") else c for c in command]
                command = [s.replace("~",home_dir) for s in command]

        elif type(command) == str or type(command) == unicode and command.startswith("python"):
            if not self.local_process:
                command = command.replace("python", self.get_extended_command(get_artemis_config_value(section=self.get_ip(), option="python")), 1)
            else:
                command = command.replace("python", sys.executable)
            command = command.replace("~",home_dir)
        else:
            raise NotImplementedError()
        return command

    def get_extended_command(self,command):
        return "echo $$ ; exec %s"%command

    def deconstruct(self, message=signal.SIGINT):
        '''
        This completely and safely deconstructs a remote connection. It will also be called at program shutdown.
        kills itself if alive, then closes remote connection if applicable
        :return:
        '''
        print("Deconstructing {}".format(self.name))
        if not self.cp_started:
            # Nothing has happened yet
            return

        if self.is_alive():
            self.kill(signal=message)
            time.sleep(4.0)
        if self.is_alive():
            self.kill(signal.SIGTERM)
        if not self.is_local():
            self.ssh_conn.close()

    def is_local(self):
        return self.local_process

    def get_name(self):
        return self.name

    def get_ip(self):
        return self.ip_address

    def get_id(self):
        return self.ip_address + "_" + str(self.id)

    def _assign_pid(self,pid):
        self.pid = pid

    def get_pid(self):
        assert self.cp_started, "This ChildProcess has not been executed yet, no pid available yet. run execute_child_process() first"
        return self.pid

    def get_ssh_connection(self):
        assert not self.is_local(), "A local ChildProcess does not have a ssh_connection"
        assert self.cp_started, "A ssh_connection will only be available after running execute_child_process()"
        return self.ssh_conn

    def execute_child_process(self, get_pty = False):
        '''
        Executes ChildProcess in a non-blocking manner. This returns immediately.
        This method returns a tuple (stdin, stdout, stderr) of the child process
        :return:
        '''
        if not self.is_local():
            self.ssh_conn = get_ssh_connection(self.ip_address)
        command = self.prepare_command(self.command)
        pid, stdin, stdout, stderr = self._run_command(command, get_pty=False)
        self._assign_pid(pid)
        self.cp_started = True
        return (stdin, stdout, stderr)

    def _run_command(self,command, get_pty=False):
        '''
        execute the given command
        :param command: string, to execute
        :return: (stdin, stdout, stderr)
        '''
        if self.local_process:
            if type(command) == list:
                # print (" ".join(c for c in command))
                sub = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            elif type(command) == str or type(command) == unicode:
                shlexed_command = shlex.split(command)
                sub = subprocess.Popen(shlexed_command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                # sub.communicate()
            else:
                raise NotImplementedError()
            self.sub = sub
            stdin = sub.stdin
            stdout = sub.stdout
            stderr = sub.stderr
            pid = sub.pid
        else:
            stdin, stdout, stderr = self.ssh_conn.exec_command(command)
            pid = stdout.readline().strip()
        return (pid, stdin, stdout, stderr)

    def kill(self, signal=signal.SIGINT):
        '''
        Kills the process by
        (remote) sending 'kill -s signal pid' to the server.
        (local)  os.kill(pid, signal)
        default signal is SIGINT
        This call does not block. The success of killing the process
        needs to be determined by the user. E.g by calling is_alive()
        :return:
        '''

        if not self.cp_started:
            print("Not started yet, no kill command will be sent")
            return
        if self.is_local():
            self.sub.send_signal(signal)
        else:
            kill_command = "kill -s %s %s" %(signal, self.get_pid())
            self._run_command(kill_command)

    def is_alive(self):
        if not self.cp_started:
            return False
        if self.is_local():
            return self.sub.poll() == None
        else:
            command = "echo $$; exec ps -h -p %s"%self.get_pid()
            _,_,stdout,_ = self._run_command(command)
            return self.get_pid() in stdout.read()


def get_ssh_connection(ip_address):
    '''
    This returns a ssh_connection to the given ip_address. Make sure to close the connection afterwards.
    Requires a public/private key to be set up with the remote system. The location of the private key can be
    specified in .artemisrc or, if not specified, will be looked for in ~/.ssh/id_rsa
    :param ip_address:
    :return:
    '''

    try:
        path_to_private_key = get_artemis_config_value(section=ip_address, option="private_key")
    except NoOptionError:
        path_to_private_key = os.path.join(os.path.expanduser("~"),".ssh/id_rsa")

    private_key = paramiko.RSAKey.from_private_key_file(os.path.expanduser(path_to_private_key))
    username = get_artemis_config_value(section=ip_address, option="username")
    ssh_conn = paramiko.SSHClient()
    ssh_conn.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh_conn.connect(hostname=ip_address, username=username, pkey=private_key)
    return ssh_conn


def execute_command(ip_address, command, blocking=True):
    '''
    This method spawns a child-process (either locally or remote, depending on the ip_address). Then it executes the given command and handles communication.
    If blocking is True, then this call does not return before the child process returns. It then prints stdout to console, followed by stderr.
    If blocking is False, then this call returns immediately and asynchronously forwards stdout and stderr to the console in separate threads.
    If ip_address is local, then the command will be split using shlex.split() and must be formatted accordingly. The subprocess call is executed with shell=False
    for all the right reasons.
    :param ip_address:
    :param command: String. command to execute
    :param blocking:
    :return:
    '''
    if ip_address in get_local_ips():

        sub = subprocess.Popen(shlex.split(command),stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)
        stdout_pipe = sub.stdout
        stderr_pipe = sub.stderr

    else:
        ssh_conn = get_ssh_connection(ip_address)
        transport = ssh_conn.get_transport()
        channel = transport.open_session()
        channel.exec_command(command)#
        bufsize=-1
        stdout_pipe = channel.makefile('r', bufsize)
        stderr_pipe = channel.makefile_stderr('r',bufsize)

    #stdout
    t1 = ParamikoPrintThread(source_pipe=stdout_pipe, target_pipe=sys.stdout)
    t1.start()
    # stderr
    t2 = ParamikoPrintThread(source_pipe=stderr_pipe, target_pipe=sys.stderr)
    t2.start()
    if blocking:
        t1.join()
        t2.join()


def check_ssh_connection(ip_address):
    '''
    tries to load necessary information from the ~/.artemisrc file and execute a test remote function call. This is to verify that ssh connection is available
    :param ip_address: the ip_address to call against
    :return:
    '''

    test_function = 'python -c "import socket; print([l for l in ([ip for ip in socket.gethostbyname_ex(socket.gethostname())[2] if not ip.startswith(\'127.\')][:1], [[(s.connect((\'8.8.8.8\', 53)), s.getsockname()[0], s.close()) for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]][0][1]]) if l][0][0])"'
    ssh_conn = get_ssh_connection(ip_address=ip_address)
    stdin , stdout, stderr = ssh_conn.exec_command(test_function)
    out = stdout.read().strip()
    assert out == ip_address, "The remote server resolved a different ip-address than the one this computer used to contact it. This must not be a problem, but may be worth investigating"
    err = stderr.read()
    assert not err, "The remote server could not execute the test function. It returned the following error: \n %s"%err
    ssh_conn.close()


def check_if_port_is_free(ip_address, port):
    '''
    This checks if the remote server is able to accept requests at the given port.
    :param connections: A list ["ip_address:port"] of strings
    :param force_close: In case the port is in use, force close the program that is occupying it.
    :return:
    '''

    try:
        port = int(port)
    except ValueError:
        print ("Please provide a valid port for address %s. Received %s instead" %(ip_address, port))
        raise

    if ip_address in get_local_ips():
        import socket
        s=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.bind((ip_address,port))
        except:
            raise
        finally:
            s.close()
    else:
        check_ssh_connection(ip_address)
        ssh_connect = get_ssh_connection(ip_address=ip_address)
        check_port_function = 'python -c "import socket; s=socket.socket(socket.AF_INET, socket.SOCK_STREAM);s.bind((\'%s\',%i));s.close()"'%(ip_address,port)
        stdin , stdout, stderr = ssh_connect.exec_command(check_port_function)
        err = stderr.read()
        assert not err, "The remote address %s cannot allocate port %i. The following error was raised: \n %s" % (ip_address, port,err.strip().split("\n")[-1])



from artemis.fileman.primitive_specifiers import PrimativeSpecifier, load_primative

__author__ = 'peter'

def test_primative_specifiers():

    class MyThing(PrimativeSpecifier):

        def __init__(self, a, b, c):
            self.a=a
            self.b=b
            self.c=c

    thing1 = MyThing(1, 2, 3)

    primative = thing1.to_primative()

    thing2 = load_primative(primative)
    assert thing2.a==1
    assert thing2.b==2
    assert thing2.c==3


if __name__ == '__main__':
    test_primative_specifiers()



import time
from artemis.fileman.disk_memoize import memoize_to_disk, clear_memo_files_for_function, DisableMemos, memoize_to_disk_and_cache, \
    memoize_to_disk_test, memoize_to_disk_and_cache_test
from artemis.general.test_mode import set_test_mode
import numpy as np
from pytest import raises

__author__ = 'peter'

@memoize_to_disk_test
def compute_slow_thing(a, b, c):
    call_time = time.time()
    time.sleep(0.01)
    return (a+b)/float(c), call_time


def test_memoize_to_disk():

    clear_memo_files_for_function(compute_slow_thing)

    t = time.time()
    num, t1 = compute_slow_thing(1, 3, 3)
    assert t-t1 < 0.01
    assert num == (1+3)/3.

    num, t2 = compute_slow_thing(1, 3, 3)
    assert num == (1+3)/3.
    assert t2==t1

    num, t3 = compute_slow_thing(1, 3, 4)
    assert num == (1+3)/4.
    assert t3 != t1

    with DisableMemos():
        num, t4 = compute_slow_thing(1, 3, 4)
        assert num == (1+3)/4.
        assert t4 != t3

    num, t5 = compute_slow_thing(1, 3, 4)
    assert num == (1+3)/4.
    assert t5 == t3


@memoize_to_disk_and_cache_test
def complex_arg_fcn(a, b):
    time.sleep(0.01)
    return str(a)+str(b), time.time()


def test_complex_args():
    """
    Make sure that we detect changes in nested arguments.
    :return:
    """
    arr = np.random.rand(3, 4)
    clear_memo_files_for_function(complex_arg_fcn)
    t = time.time()
    out1, t1 = complex_arg_fcn([1, 2], b={'a': 3, 'b': arr})
    assert t-t1 < 0.02
    out2, t2 = complex_arg_fcn([1, 2], b={'a': 3, 'b': arr})
    assert out1==out2
    assert t1==t2
    arr[2, 3] = -1
    out3, t3 = complex_arg_fcn([1, 2], b={'a': 3, 'b': arr})
    assert out1!=out3
    assert t1!=t3


@memoize_to_disk_and_cache_test
def compute_slow_thing_again(a, b, c):
    call_time = time.time()
    time.sleep(0.01)
    return (a+b)/float(c), call_time


def test_memoize_to_disk_and_cache():

    clear_memo_files_for_function(compute_slow_thing_again)

    t = time.time()
    num1, t1 = compute_slow_thing_again(1, 3, 3)
    assert t-t1 < 0.01
    assert num1 == (1+3)/3.

    num2, t2 = compute_slow_thing_again(1, 3, 3)
    assert num2 == (1+3)/3.
    assert num2 is num1
    assert t2==t1

    num, t3 = compute_slow_thing_again(1, 3, 4)
    assert num == (1+3)/4.
    assert t3 != t1

    with DisableMemos():
        num, t4 = compute_slow_thing_again(1, 3, 4)
        assert num == (1+3)/4.
        assert t4 != t3

    num, t5 = compute_slow_thing_again(1, 3, 4)
    assert num == (1+3)/4.
    assert t5 == t3


def test_clear_error_for_missing_arg():

    clear_memo_files_for_function(compute_slow_thing)

    with raises(AssertionError):
        compute_slow_thing(1)


def test_clear_arror_for_wrong_arg():

    clear_memo_files_for_function(compute_slow_thing)

    with raises(AssertionError):
        compute_slow_thing(a=1, b=2, c=3, d=4)


@memoize_to_disk_test
def compute_slow_thing_with_kwargs(a, **kwargs):
    call_time = time.time()
    time.sleep(0.01)
    return (a+kwargs['b'])/float(kwargs['c']), call_time


def test_unnoticed_wrong_arg_bug_is_dead():

    clear_memo_files_for_function(compute_slow_thing)
    compute_slow_thing(a=1, b=2, c=3)  # Creates a memo
    with raises(AssertionError):
        compute_slow_thing(a=1, b=2, see=3)  # Previously, this was not caught, leading you not to notice your typo


def test_catch_kwarg_error():

    clear_memo_files_for_function(compute_slow_thing_with_kwargs)

    t = time.time()
    num, t1 = compute_slow_thing_with_kwargs(1, b=2, c=4)
    assert num == (1+2)/4.
    assert t1 > t

    num, t2 = compute_slow_thing_with_kwargs(1, b=2, c=6)
    assert num == (1+2)/6.
    assert t2>t1

    num, t3 = compute_slow_thing_with_kwargs(1, b=2, c=4)
    assert num == (1+2)/4.
    assert t3 == t1


if __name__ == '__main__':
    set_test_mode(True)
    test_unnoticed_wrong_arg_bug_is_dead()
    test_catch_kwarg_error()
    test_clear_arror_for_wrong_arg()
    test_clear_error_for_missing_arg()
    test_memoize_to_disk_and_cache()
    test_memoize_to_disk()
    test_complex_args()



import numpy as np

from artemis.ml.datasets.datasets import DataCollection, DataSet


__author__ = 'peter'


def get_synthetic_clusters_dataset(n_clusters = 4, n_dims = 20, n_training = 1000, n_test = 200,
        sparsity = 0.5, flip_noise = 0.1, seed = 3425, dtype = 'float32'):
    """
    A dataset consisting of clustered binary data with "bit-flip" noise, and the corresponding cluster labels.
    This should be trivially solvable by any classifier, and serves as a basic test of whether your classifier is
    completely broken or not.

    :param n_clusters:
    :param n_dims:
    :param n_samples_training:
    :param n_samples_test:
    :param sparsity:
    :param flip_noise:
    :param seed:
    :return:
    """

    rng = np.random.RandomState(seed)
    labels = rng.randint(n_clusters, size = n_training+n_test)  # (n_samples, )
    centers = rng.rand(n_clusters, n_dims) < sparsity  # (n_samples, n_dims)
    input_data = centers[labels]
    input_data = np.bitwise_xor(input_data, rng.rand(*input_data.shape) < flip_noise).astype(dtype)

    return DataSet(
        training_set = DataCollection(input_data[:n_training], labels[:n_training]),
        test_set = DataCollection(input_data[n_training:], labels[n_training:]),
        name = 'Synthetic Clusters Dataset'
        )



import numpy as np

from artemis.general.should_be_builtins import bad_value, all_equal
from artemis.ml.tools.processors import OneHotEncoding


__author__ = 'peter'


class DataSet(object):

    def __init__(self, training_set, test_set, validation_set = None, name = None):

        sets = [training_set, test_set] + [validation_set] if validation_set is not None else []
        assert all_equal([[x.shape[1:] for x in s.inputs] for s in sets])
        assert all_equal([[x.shape[1:] for x in s.targets] for s in sets])
        self.training_set = training_set
        self.test_set = test_set
        self._validation_set = validation_set
        self._name = name
        self._n_categories = None

    @property
    def validation_set(self):
        if self._validation_set is None:
            raise Exception('Validation set does not exist')
        else:
            return self._validation_set

    @property
    def input_shapes(self):
        return [x.shape[1:] for x in self.training_set.inputs]

    @property
    def input_shape(self):
        return self.training_set.input.shape[1:]

    @property
    def input_size(self):
        return np.prod(self.input_shape)

    @property
    def target_shapes(self):
        return [x.shape[1:] for x in self.training_set.targets]

    @property
    def target_shape(self):
        return self.training_set.target.shape[1:]

    @property
    def target_size(self):
        return np.prod(self.target_shape, dtype=np.int)

    @property
    def name(self):
        return self.__class__.__name__ if self._name is None else self._name

    @property
    def n_categories(self):
        if self._n_categories is None:
            assert self.training_set.target.dtype in (int, str, np.int32), \
                'n_categories is only a valid attribute when target data is int or str.  It is %s' \
                % (self.training_set.target.dtype, )
            self._n_categories = len(np.unique(self.training_set.target))
        return self._n_categories  # TODO: Do this properly (assert that train/test have same categories, etc

    @property
    def xyxy(self):
        """
        Shorthand to return (input_of_training_set, targets_of_training_set, input_of_test_set, targets_of_test_set)
        :return:
        """
        return self.training_set.input, self.training_set.target, self.test_set.input, self.test_set.target

    def process_with(self, inputs_processor=None, targets_processor = None):
        return DataSet(
            training_set=self.training_set.process_with(inputs_processor, targets_processor),
            test_set=self.test_set.process_with(inputs_processor, targets_processor),
            validation_set=self._validation_set.process_with(inputs_processor, targets_processor) if self._validation_set is not None else None,
        )

    @staticmethod
    def from_xyxy(training_inputs, training_targets, test_inputs, test_targets):
        return DataSet(training_set = DataCollection(training_inputs, training_targets), test_set = DataCollection(test_inputs, test_targets))

    @staticmethod
    def from_xy(x, y, training_fraction, validation_fraction = None):
        return DataCollection(x, y).to_dataset(training_fraction=training_fraction, validation_fraction=validation_fraction)

    def __repr__(self):
        return '<%s with %s training samples, %s test samples, input_shapes = %s, target_shapes = %s at %s>' \
            % (self.name, self.training_set.n_samples, self.test_set.n_samples,
               [x.shape[1:] for x in self.training_set.inputs], [x.shape[1:] for x in self.training_set.targets],
                hex(id(self))
            )

    def shorten(self, n_samples):
        """
        Shorten the training/test sets to n_samples each.  This is useful in code tests, when we just
        want a little bit of the dataset to make sure that the code runs.
        """
        return DataSet(training_set=self.training_set.shorten(n_samples), test_set=self.test_set.shorten(n_samples),
            validation_set=self._validation_set.shorten(n_samples) if self._validation_set is not None else None)

    def to_onehot(self, form = 'bin', n_categories = None, dtype = None):
        if n_categories is None:
            n_categories = self.n_categories  # Will throw an exception if not a categorical target
        encoder = OneHotEncoding(n_categories, form=form, dtype=dtype)
        return self.process_with(targets_processor=lambda (t, ): (encoder(t), ))


class DataCollection(object):

    def __init__(self, inputs, targets):
        if not isinstance(inputs, (list, tuple)):
            inputs = (inputs, )
        if not isinstance(targets, (list, tuple)):
            targets = (targets, )
        n_samples = inputs[0].shape[0]
        assert all(n_samples == d.shape[0] for d in inputs) and all(n_samples == l.shape[0] for l in targets)
        self._inputs = inputs
        self._targets = targets
        self._n_samples = n_samples

    @property
    def n_samples(self):
        return self._n_samples

    @property
    def inputs(self):
        return self._inputs

    @property
    def targets(self):
        return self._targets

    @property
    def input(self):
        """Deprecated.  Use self.x"""
        only_input, = self._inputs
        return only_input

    @property
    def x(self):
        return self.input

    @property
    def target(self):
        """Deprecated.  Use self.y"""
        only_target, = self._targets
        return only_target

    @property
    def y(self):
        return self.target

    def minibatch_iterator(self, **kwargs):
        """
        See minibatch_iterator
        """
        return minibatch_iterator(**kwargs)(self)

    def process_with(self, inputs_processor=None, targets_processor = None):
        inputs = inputs_processor(self._inputs) if inputs_processor is not None else self._inputs
        targets = targets_processor(self._targets) if targets_processor is not None else self._targets
        return DataCollection(inputs, targets)

    def shorten(self, n_samples):
        new_inputs = [x[:n_samples] for x in self.inputs]
        new_targets = [x[:n_samples] for x in self.targets]
        return DataCollection(new_inputs, new_targets)

    def to_dataset(self, training_fraction, validation_fraction = None):
        assert 0<training_fraction<1
        assert validation_fraction is None or 0<validation_fraction<1
        training_end = int(training_fraction*self.n_samples+.5)
        test_start = self.n_samples - training_end if validation_fraction is None else self.n_samples - int(validation_fraction*self.n_samples+.5)
        return DataSet(
            training_set=DataCollection(inputs = [x[:training_end] for x in self.inputs], targets=[y[:training_end] for y in self.targets]),
            test_set=DataCollection(inputs = [x[test_start:] for x in self.inputs], targets=[y[test_start:] for y in self.targets]),
            validation_set=DataCollection(inputs = [x[training_end:test_start] for x in self.inputs], targets=[y[training_end:test_start] for y in self.targets]) if validation_fraction is not None else None,
            )

    @property
    def xy(self):
        return self.input, self.target


def minibatch_iterator(minibatch_size = 1, epochs = 1, final_treatment = 'stop', single_channel = False):
    """
    :param minibatch_size:
    :param epochs:
    :param final_treatment:
    :param single_channel:
    :return: A function that, when called with a Data Collection, returns an iterator.
    """

    def iterator(data_collection):
        """
        :param data_collection: A DataCollection object
        :yield: A 2-tuple of (input_data, label_data)
        """
        assert isinstance(data_collection, DataCollection)
        i = 0
        n_samples = data_collection.n_samples
        total_samples = epochs * n_samples

        true_minibatch_size = n_samples if minibatch_size == 'full' else \
            minibatch_size if isinstance(minibatch_size, int) else \
            bad_value(minibatch_size)

        if single_channel:
            input_data = data_collection.input
            target_data = data_collection.target
        else:
            input_data = data_collection.inputs
            target_data = data_collection.targets

        while i < total_samples:
            next_i = i + true_minibatch_size
            segment = np.arange(i, next_i) % n_samples
            if next_i > total_samples:
                if final_treatment == 'stop':
                    break
                elif final_treatment == 'truncate':
                    next_i = total_samples
                else:
                    raise Exception('Unknown final treatment: %s' % final_treatment)
            if single_channel:
                input_minibatch = input_data[segment]
                target_minibatch = target_data[segment]
            else:
                input_minibatch, = [d[segment] for d in input_data]
                target_minibatch, = [d[segment] for d in target_data]

            yield next_i, input_minibatch, target_minibatch
            i = next_i

    return iterator



import re
__author__ = 'peter'


class ReDict(dict):
    """
    A dictionary where keys are regular expressions, which try to match the index you reference with.

    If the number of keys matching your index does not equal 1 (0 or >=2), you get a KeyError.

    See test_regexp_dict for examples, and documentation for python's "re" module for info on regular expressions.
    """

    def __init__(self, dict_initializer):
        """
        dict_initializer is any argument that you'd give to a dictionary with the exception that keys must be strings
            or None.  None in this case means default - in case nothing else matches.
        """
        dict.__init__(self, dict_initializer)
        assert all(isinstance(k, basestring) or k is None for k in self), 'All keys to a Redict must be strings or None'

    def __getitem__(self, index):
        match_found = False
        if index is None:
            return dict.__getitem__(self, None)
        for k, v in self.iteritems():
            if k is not None and re.match(k, index) is not None:
                if match_found:
                    raise MultipleMatchesError('Multiple Matches to expression %s: %s.  If this is what you want use get_matches'
                        % (index, self.get_matches(index)))
                match_found = True
                value = v
        if not match_found:
            if None in self:
                value = self[None]
            else:
                raise KeyError('No matches to expression %s' % (index, ))
        return value

    def __contains__(self, index):

        return (None in self.viewkeys()) if index is None else any(re.match(k, index) for k in self if k is not None)

    def get_matches(self, index, as_redict = True):
        """
        Get matching keys, NOT including the default None key.
        """
        matching_dict = {k: v for k, v in self.iteritems() if k is not None and re.match(k, index) is not None}
        return ReDict(matching_dict) if as_redict else matching_dict

    def get(self, item, default):
        try:
            return self[item]
        except KeyError:
            return default


class MultipleMatchesError(Exception):
    pass


class ReCurseDict(ReDict):
    """
    A recursively nested ReDict.  You can build this out of nested dictionaries, so that when you reference
    an item that is another dictionary, it will reference that inner dictionary to find a match.
    """

    def __init__(self, dict_initialzer):
        """
        :param dict_initialzer: Don't try to be funny here an give dicts containing themselves.
        """
        ReDict.__init__(self, dict_initialzer)
        for k, v in self.iteritems():
            if isinstance(v, dict):
                self[k] = ReCurseDict(v)

    def __getitem__(self, item):
        result = ReDict.__getitem__(self, item)
        return result[item] if isinstance(result, ReCurseDict) else result



import shlex
from collections import OrderedDict
from functools import partial
from artemis.experiments.experiment_management import pull_experiments, select_experiments, select_experiment_records, \
    select_experiment_records_from_list, interpret_numbers, run_experiment_ignoring_errors, run_multiple_experiments
from artemis.experiments.experiment_record import get_all_record_ids, clear_experiment_records, \
    experiment_id_to_record_ids, load_experiment_record, ExpInfoFields, NoSavedResultError
from artemis.experiments.experiment_record_view import get_record_full_string, get_record_invalid_arg_string, \
    print_experiment_record_argtable, compare_experiment_results, show_experiment_records, get_oneline_result_string, \
    display_experiment_record
from artemis.experiments.experiments import GLOBAL_EXPERIMENT_LIBRARY, load_experiment, is_experiment_loadable
from artemis.general.display import IndentPrint, side_by_side
from artemis.general.should_be_builtins import all_equal
from tabulate import tabulate


def _setup_input_memory():
    try:
        import readline  # Makes raw_input behave like interactive shell.
        # http://stackoverflow.com/questions/15416054/command-line-in-python-with-history
    except:
        pass  # readline not available


def _warn_with_prompt(message= None, prompt = 'Press Enter to continue', use_prompt=True):
    if message is not None:
        print message
    if use_prompt:
        return raw_input('({}) >> '.format(prompt))


def browse_experiments(command=None, **kwargs):
    """
    Browse Experiments

    :param root_experiment: Optionally, the root experiment to look at.
    :param catch_errors: True if you want to catch any errors here
    :param close_after: Close this menu after running an experiment
    :param just_last_record: Just show the last record for the experiment
    """
    browser = ExperimentBrowser(**kwargs)
    browser.launch(command=command)


class ExperimentBrowser(object):

    QUIT = 'Quit'
    HELP_TEXT = """
This program lists the experiments that you have defined (referenced by E#) alongside the records of console output,
plots, results, referenced by (E#.R# - for example 4.1) created by running these experiments.  Command examples:

> 4                   Run experiment 4
> run 4               Run experiment 4
> run 4-6             Run experiment 4, 5, and 6
> run all             Run all experiments
> run unfinished      Run all experiments which have not yet run to completion.
> run 4-6 -s          Run experiments 4, 5, and 6 in sequence, and catch all errors.
> run 4-6 -e          Run experiments 4, 5, and 6 in sequence, and stop on errors
> run 4-6 -p          Run experiments 4, 5, and 6 in parallel processes, and catch all errors.
> run 4-6 -p2         Run experiments 4, 5, and 6 in parallel processes, using up to 2 processes at a time.
> call 4              Call experiment 4 (like running, but doesn't save a record)
> filter 4-6          Just show experiments 4-6 and their records
> filter has:xyz      Just show experiments with "xyz" in the name and their records
> filter 1diff:3      Just show all experiments that have no more than 1 argument different from experiment 3.
> filter -            Clear the filter and show the full list of experiments
> results 4-6         View the results experiments 4, 5, 6
> view results        View just the columns for experiment name and result
> view full           View all columns (the default view)
> show 4              Show the output from the last run of experiment 4 (if it has been run already).
> show 4-6            Show the output of experiments 4,5,6 together.
> records             Browse through all experiment records.
> allruns             Toggle between showing all past runs of each experiment, and just the last one.
> compare 4.1,5.3     Print a table comparing the arguments and results of records 4.1 and 5.3.
> selectexp 4-6       Show the list of experiments (and their records) selected by the "experiment selector" "4-6" (see below for possible experiment selectors)
> selectrec 4-6       Show the list of records selected by the "record selector" "4-6" (see below for possible record selectors)
> sidebyside 4.1,5.3  Display the output of record from experiments 4.1,5.3 side by side.
> delete 4.1          Delete record 1 of experiment 4
> delete unfinished   Delete all experiment records that have not run to completion
> delete 4-6          Delete all records from experiments 4, 5, 6.  You will be asked to confirm the deletion.
> pull 1-4 machine1   Pulls records from experiments 1,2,3,4 from machine1 (requires some setup, see artemis.remote.remote_machines.py)
> q                   Quit.
> r                   Refresh list of experiments.

Commands 'run', 'call', 'filter', 'pull', '1diff', 'selectexp' allow you to select experiments.  You can select
experiments in the following ways:

    Experiment
    Selector        Action
    --------------- ---------------------------------------------------------------------------------------------------
    4               Select experiment #4
    4-6             Select experiments 4, 5, 6
    all             Select all experiments
    unfinished      Select all experiment for which there are no records of it being run to completion.
    invalid         Select all experiments where all records were made before arguments to the experiment have changed
    has:xyz         Select all experiments with the string "xyz" in their names
    hasnot:xyz      Select all experiments without substring "xyz" in their names
    1diff:3         Select all experiments who have no more than 1 argument which is different from experiment 3's arguments.

Commands 'results', 'show', 'records', 'compare', 'sidebyside', 'selectrec', 'delete' allow you to specify a range of
experiment records.  You can specify records in the following ways:

    Record
    Selector        Action
    --------------- ---------------------------------------------------------------------------------------------------
    4.2             Select record 2 for experiment 4
    4               Select all records for experiment 4
    4-6             Select all records for experiments 4, 5, 6
    4.2-5           Select records 2, 3, 4, 5 for experiment 4
    4.3,4.4         Select records 4.3, 4.4
    all             Select all records
    old             Select all records that are not the the most recent run for that experiment
    unfinished      Select all records that have not run to completion
    invalid         Select all records for which the arguments to their experiments have changed since they were run
    errors          Select all records that ended in error
    invalid|errors  Select all records that are invalid or ended in error (the '|' can be used to "or" any of the above)
    invalid&errors  Select all records that are invalid and ended in error (the '&' can be used to "and" any of the above)
"""

    def __init__(self, root_experiment = None, catch_errors = False, close_after = False, just_last_record = False,
            view_mode ='full', raise_display_errors=False, run_args=None, keep_record=True, truncate_result_to=100):

        if run_args is None:
            run_args = {}
        if 'keep_record' not in run_args:
            run_args['keep_record'] = keep_record
        self.root_experiment = root_experiment
        self.close_after = close_after
        self.just_last_record = just_last_record
        self.catch_errors = catch_errors
        self.exp_record_dict = self.reload_record_dict()
        self.raise_display_errors = raise_display_errors
        self.view_mode = view_mode
        self._filter = None
        self.run_args = {} if run_args is None else run_args
        self.truncate_result_to = truncate_result_to

    def reload_record_dict(self):

        names = GLOBAL_EXPERIMENT_LIBRARY.keys()

        if self.root_experiment is not None:
            # We could just go [ex.name for ex in self.root_experiment.get_all_variants(include_self=True)]
            # but we want to preserve the order in which experiments were created
            descendents_of_root = set(ex.name for ex in self.root_experiment.get_all_variants(include_self=True))
            names = [name for name in names if name in descendents_of_root]

        d= OrderedDict((name, experiment_id_to_record_ids(name)) for name in names)
        if self.just_last_record:
            for k in d.keys():
                d[k] = [d[k][-1]] if len(d[k])>0 else []
        return d

    def launch(self, command=None):

        func_dict = {
            'run': self.run,
            'test': self.test,
            'show': self.show,
            'call': self.call,
            'selectexp': self.selectexp,
            'selectrec': self.selectrec,
            'allruns': self.allruns,
            'view': self.view,
            'h': self.help,
            'filter': self.filter,
            'explist': self.explist,
            'sidebyside': self.side_by_side,
            'argtable': self.argtable,
            'compare': self.compare,
            'display': self.display,
            'delete': self.delete,
            'errortrace': self.errortrace,
            'q': self.quit,
            'records': self.records,
            'pull': self.pull
            }

        while True:
            all_experiments = self.reload_record_dict()

            print "==================== Experiments ===================="
            self.exp_record_dict = all_experiments if self._filter is None else \
                OrderedDict((exp_name, all_experiments[exp_name]) for exp_name in select_experiments(self._filter, all_experiments))
            print self.get_experiment_list_str(self.exp_record_dict, just_last_record=self.just_last_record, view_mode=self.view_mode, raise_display_errors=self.raise_display_errors, truncate_result_to=self.truncate_result_to)
            if self._filter is not None:
                print '[Filtered with "{}" to show {}/{} experiments]'.format(self._filter, len(self.exp_record_dict), len(all_experiments))
            print '-----------------------------------------------------'
            if command is None:
                user_input = raw_input('Enter command or experiment # to run (h for help) >> ').lstrip(' ').rstrip(' ')
            else:
                user_input=command
                command = None
            with IndentPrint():
                try:
                    split = user_input.split(' ')
                    if len(split)==0:
                        continue
                    cmd = split[0]
                    args = split[1:]

                    if cmd == '':
                        continue
                    elif cmd in func_dict:
                        out = func_dict[cmd](*args)
                    elif interpret_numbers(cmd) is not None:
                        if not any(x in args for x in ('-s', '-e', '-p')):
                            args = args + ['-e']
                        out = self.run(cmd, *args)
                    elif cmd == 'r':  # Refresh
                        continue
                    else:
                        response = raw_input('Unrecognised command: "{}".  Type "h" for help or Enter to continue. >'.format(cmd))
                        if response.lower()=='h':
                            self.help()
                        out = None
                    if out is self.QUIT or self.close_after:
                        break
                except Exception as name:
                    if self.catch_errors:
                        res = raw_input('%s: %s\nEnter "e" to view the stacktrace, or anything else to continue.' % (name.__class__.__name__, name.message))
                        if res == 'e':
                            raise
                    else:
                        raise

    @staticmethod
    def get_experiment_list_str(exp_record_dict, just_last_record, view_mode='full', raise_display_errors=False, truncate_result_to=100):

        headers = {
            'full': ['E#', 'R#', 'Name', 'Last Run' if just_last_record else 'All Runs', 'Duration', 'Status', 'Valid', 'Result'],
            'results': ['E#', 'R#', 'Name', 'Result']
            }[view_mode]

        rows = []

        def get_field(header):
            try:
                return \
                    index if header=='#' else \
                    (str(i) if j==0 else '') if header == 'E#' else \
                    j if header == 'R#' else \
                    (name if j==0 else '') if header=='Name' else \
                    experiment_record.info.get_field_text(ExpInfoFields.TIMESTAMP) if header in ('Last Run', 'All Runs') else \
                    experiment_record.info.get_field_text(ExpInfoFields.RUNTIME) if header=='Duration' else \
                    experiment_record.info.get_field_text(ExpInfoFields.STATUS) if header=='Status' else \
                    get_record_invalid_arg_string(experiment_record) if header=='Valid' else \
                    get_oneline_result_string(experiment_record, truncate_to=truncate_result_to) if header=='Result' else \
                    '???'
            except:
                if raise_display_errors:
                    raise
                return '<Display Error>'

        for i, (exp_id, record_ids) in enumerate(exp_record_dict.iteritems()):
            if len(record_ids)==0:
                if exp_id in exp_record_dict:
                    rows.append([str(i), '', exp_id, '<No Records>', '-', '-', '-', '-'])
            else:
                for j, record_id in enumerate(record_ids):
                    index, name = ['{}.{}'.format(i, j), exp_id] if j==0 else ['{}.{}'.format('`'*len(str(i)), j), exp_id]
                    try:
                        experiment_record = load_experiment_record(record_id)
                    except:
                        experiment_record = None
                    rows.append([get_field(h) for h in headers])
        assert all_equal([len(headers)] + [len(row) for row in rows]), 'Header length: {}, Row Lengths: \n  {}'.format(len(headers), '\n'.join([len(row) for row in rows]))
        table = tabulate(rows, headers=headers)
        return table

    def run(self, user_range, mode='-s', raise_exceptions = ''):
        assert mode in ('-s', '-e') or mode.startswith('-p')
        ids = select_experiments(user_range, self.exp_record_dict)
        run_multiple_experiments(
            experiments=[load_experiment(eid) for eid in ids],
            parallel=len(ids)>1 and mode.startswith('-p'),
            raise_exceptions = raise_exceptions=='-e',
            run_args=self.run_args
            )

        result = _warn_with_prompt('Finished running {} experiment{}.'.format(len(ids), '' if len(ids)==1 else 's'),
                use_prompt=not self.close_after,
                prompt='Press Enter to Continue, or "q" then Enter to Quit')
        if result=='q':
            quit()

    def test(self, user_range):
        ids = select_experiments(user_range, self.exp_record_dict)
        for experiment_identifier in ids:
            load_experiment(experiment_identifier).test()

    def help(self):
        _warn_with_prompt(self.HELP_TEXT, prompt = 'Press Enter to exit help.', use_prompt=not self.close_after)

    def show(self, user_range, parallel_arg = None):
        """
        :param user_range:  A range specifying the record
        :param parallel_arg: -p to print logs side-by-side, and -s to print them in sequence.
        """
        records = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        parallel_text = {'-p': True, '-s': False, None: None}[parallel_arg]
        has_matplotlib_figures=show_experiment_records(records, parallel_text = parallel_text)
        if has_matplotlib_figures:
            from matplotlib import pyplot as plt
            print '\n\n... Close all figures to return to experiment browser ...'
            plt.ioff()
            plt.show()
        else:
            _warn_with_prompt(use_prompt=not self.close_after)

    def display(self, user_range = 'all'):
        records = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        with IndentPrint("Results:", show_line=True, show_end=True):
            for record in records:
                with IndentPrint(record.get_id(), show_line=True, show_end=True):
                    display_experiment_record(record)

        _warn_with_prompt(use_prompt=not self.close_after)

    def errortrace(self, user_range):
        records = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        with IndentPrint("Error Traces:", show_line=True, show_end=True):
            for record in records:
                with IndentPrint(record.get_id(), show_line=True):
                    print record.get_error_trace()
        _warn_with_prompt(use_prompt=not self.close_after)

    def delete(self, user_range):
        records = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        print '{} out of {} Records will be deleted.'.format(len(records), sum(len(recs) for recs in self.exp_record_dict.values()))
        with IndentPrint():
            print ExperimentRecordBrowser.get_record_table(records, )
        response = raw_input('Type "yes" to continue. >')
        if response.lower() == 'yes':
            clear_experiment_records([record.get_id() for record in records])
            print 'Records deleted.'
        else:
            _warn_with_prompt('Records were not deleted.', use_prompt=not self.close_after)

    def call(self, user_range):
        ids = select_experiments(user_range, self.exp_record_dict)
        for experiment_identifier in ids:
            load_experiment(experiment_identifier)()

    def selectexp(self, user_range):
        exps_to_records = select_experiments(user_range, self.exp_record_dict, return_dict=True)
        with IndentPrint():
            print self.get_experiment_list_str(exps_to_records, just_last_record=self.just_last_record, view_mode=self.view_mode, raise_display_errors=self.raise_display_errors)
            # print ExperimentRecordBrowser.get_record_table(record_ids)
        _warn_with_prompt('Experiment Selection "{}" includes {} out of {} experiments.'.format(user_range, len(exps_to_records), len(self.exp_record_dict)), use_prompt=not self.close_after)

    def selectrec(self, user_range):
        records = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        with IndentPrint():
            print ExperimentRecordBrowser.get_record_table(records)
        _warn_with_prompt('Record Selection "{}" includes {} out of {} records.'.format(user_range, len(records), sum(len(recs) for recs in self.exp_record_dict.values())), use_prompt=not self.close_after)

    def side_by_side(self, user_range):
        records = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        print side_by_side([get_record_full_string(rec) for rec in records], max_linewidth=128)
        _warn_with_prompt(use_prompt=not self.close_after)

    def argtable(self, user_range):
        records = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        print_experiment_record_argtable(records)
        _warn_with_prompt(use_prompt=not self.close_after)

    def records(self, ):
        browse_experiment_records(self.exp_record_dict.keys())

    def pull(self, user_range, machine_name):
        from artemis.remote.remote_machines import get_remote_machine_info
        info = get_remote_machine_info(machine_name)
        exp_names = select_experiments(user_range, self.exp_record_dict)
        output = pull_experiments(user=info['username'], ip=info['ip'], experiment_names=exp_names, include_variants=False)
        print output

    def allruns(self, ):
        self.just_last_record = not self.just_last_record

    def filter(self, user_range):
        self._filter = user_range if user_range not in ('-', '--clear') else None

    def view(self, mode):
        self.view_mode = mode

    def compare(self, user_range):
        experiment_ids = select_experiments(user_range, self.exp_record_dict)
        experiments = [load_experiment(eid) for eid in experiment_ids]
        compare_experiment_results(experiments, error_if_no_result=True)
        _warn_with_prompt(use_prompt=not self.close_after)

    def explist(self, surround = ""):
        print "\n".join([surround+k+surround for k in self.exp_record_dict.keys()])
        _warn_with_prompt(use_prompt=not self.close_after)

    def quit(self):
        return self.QUIT


def browse_experiment_records(**kwargs):
    """
    Browse through experiment records.

    :param names: Filter by names of experiments
    :param filter_text: Filter by regular expression
    :return:
    """

    experiment_record_browser = ExperimentRecordBrowser(**kwargs)
    experiment_record_browser.launch()


class ExperimentRecordBrowser(object):

    QUIT = 'Quit'
    HELP_TEXT = """
    q:                  Quit
    r:                  Refresh
    filter <text>       filter experiments
    viewfilters         View all filters on these results
    showall:            Show all experiments ever
    allnames:           Remove any name filters
    show <number>       Show experiment with number
    side_by_side 4,6,9       Compare experiments by their numbers.
    clearall            Delete all experements from your computer
"""

    def __init__(self, experiment_names = None, filter_text = None, raise_display_errors=False):
        """
        Browse through experiment records.

        :param names: Filter by names of experiments
        :param filter_text: Filter by regular expression
        :return:
        """
        self.experiment_names = experiment_names
        self.filters = [filter_text]
        self.record_ids = self.reload_ids()
        self.raise_display_errors = raise_display_errors

    def reload_ids(self):
        return get_all_record_ids(experiment_ids= self.experiment_names, filters=self.filters)

    @staticmethod
    def get_record_table(records = None, headers = ('#', 'Identifier', 'Start Time', 'Duration', 'Status', 'Valid', 'Notes', 'Result'), raise_display_errors = False, result_truncation=100):

        d = {
            '#': lambda: i,
            'Identifier': lambda: experiment_record.get_id(),
            'Start Time': lambda: experiment_record.info.get_field_text(ExpInfoFields.TIMESTAMP, replacement_if_none='?'),
            'Duration': lambda: experiment_record.info.get_field_text(ExpInfoFields.RUNTIME, replacement_if_none='?'),
            'Status': lambda: experiment_record.info.get_field_text(ExpInfoFields.STATUS, replacement_if_none='?'),
            'Args': lambda: experiment_record.info.get_field_text(ExpInfoFields.ARGS, replacement_if_none='?'),
            'Valid': lambda: get_record_invalid_arg_string(experiment_record),
            'Notes': lambda: experiment_record.info.get_field_text(ExpInfoFields.NOTES, replacement_if_none='?'),
            'Result': lambda: get_oneline_result_string(experiment_record, truncate_to=128)
            # experiment_record.get_experiment().get_oneline_result_string(truncate_to=result_truncation) if is_experiment_loadable(experiment_record.get_experiment_id()) else '<Experiment not loaded>'
            }

        def get_col_info(headers):
            info = []
            for h in headers:
                try:
                    info.append(d[h]())
                except:
                    info.append('<Error displaying info>')
                    if raise_display_errors:
                        raise
            return info

        rows = []
        for i, experiment_record in enumerate(records):
            rows.append(get_col_info(headers))
        assert all_equal([len(headers)] + [len(row) for row in rows]), 'Header length: {}, Row Lengths: \n {}'.format(len(headers), [len(row) for row in rows])
        return tabulate(rows, headers=headers)

    def launch(self):

        func_lookup = {
            'q': self.quit,
            'h': self.help,
            'filter': self.filter,
            'showall': self.showall,
            'args': self.args,
            'rmfilters': self.rmfilters,
            'viewfilters': self.viewfilters,
            'side_by_side': self.compare,
            'show': self.show,
            'search': self.search,
            'delete': self.delete,
        }

        while True:

            print "=============== Experiment Records =================="
            # print tabulate([[i]+e.get_row() for i, e in enumerate(record_infos)], headers=['#']+_ExperimentInfo.get_headers())
            print self.get_record_table([load_experiment_record(rid) for rid in self.record_ids], raise_display_errors = self.raise_display_errors)
            print '-----------------------------------------------------'

            if self.experiment_names is not None or len(self.filters) != 0:
                print 'Not showing all experiments.  Type "showall" to see all experiments, or "viewfilters" to view current filters.'
            user_input = raw_input('Enter Command (show # to show and experiment, or h for help) >>')
            parts = shlex.split(user_input)
            if len(parts)==0:
                print "You need to specify an command.  Press h for help."
                continue
            cmd = parts[0]
            args = parts[1:]

            try:
                if cmd not in func_lookup:
                    raise _warn_with_prompt('Unknown Command: {}'.format(cmd))
                else:
                    return_val = func_lookup[cmd](*args)
                    if return_val==self.QUIT:
                        break
            except Exception as e:
                res = raw_input('%s: %s\nEnter "e" to view the stacktrace, or anything else to continue.' % (e.__class__.__name__, e.message))
                if res == 'e':
                    raise

    def _select_records(self, user_range):
        return select_experiment_records_from_list(user_range, self.record_ids)

    def quit(self):
        return self.QUIT

    def help(self):
        _warn_with_prompt(self.HELP_TEXT)

    def filter(self, filter_text):
        self.filters.append(filter_text)
        self.record_ids = self.reload_ids()

    def showall(self):
        self.filters = []
        self.experiment_names = None
        self.record_ids = self.reload_ids()

    def args(self, user_range):
        print self.get_record_table(self._select_records(user_range), headers=['Identifier', 'Args'])

    def rmfilters(self):
        self.filters = []
        self.record_ids = self.reload_ids()

    def viewfilters(self):
        _warn_with_prompt('Filtering for: \n  Names in {}\n  Expressions: {}'.format(self.experiment_names, self.filters))

    def compare(self, user_range):
        identifiers = self._select_records(user_range)
        print_experiment_record_argtable(identifiers)
        _warn_with_prompt('')

    def show(self, user_range):
        record_ids = self._select_records(user_range)
        show_experiment_records([load_experiment_record(rid) for rid in record_ids])
        _warn_with_prompt('')

    def search(self, filter_text):
        print 'Found the following Records: '
        print self.get_record_table([rid for rid in self.record_ids if filter_text in rid])
        _warn_with_prompt()

    def delete(self, user_range):
        ids = self._select_records(user_range)
        print 'We will delete the following experiments:'
        print self.get_record_table(ids)
        conf = raw_input("Going to clear {} of {} experiment records shown above.  Enter 'yes' to confirm: ".format(len(ids), len(self.record_ids)))
        if conf=='yes':
            clear_experiment_records(ids=ids)
        else:
            _warn_with_prompt("Did not delete experiments")
        self.record_ids = self.reload_ids()


if __name__ == '__main__':
    browse_experiment_records()



from artemis.general.redict import ReDict, MultipleMatchesError, ReCurseDict
import pytest

__author__ = 'peter'


def test_redict():

    d = ReDict({
        'alpha-.*': 1,
        'alpha_.*': 2,
        'beta-.*': 3,
        '.*-gamma': 4,
        'eps.*': 5,
        'phi-(eta|pi)': 6,
        'phi-zeta': 7,
        'phi-omega$': 7
        })

    assert d['alpha-beta'] == 1
    assert d['alpha_beta'] == 2
    with pytest.raises(MultipleMatchesError):
        d['alpha-gamma']  # Multiple Matches
    assert d.get_matches('alpha-gamma') == {'alpha-.*': 1, '.*-gamma': 4}
    assert d['alpha-delta'] == 1
    with pytest.raises(KeyError):
        d['delta-alpha']  # No match
    with pytest.raises(KeyError):
        d['delta_gamma']  # No match
    assert d.get_matches('delta_gamma') == {}
    assert d.get_matches('delta-alpha') == {}
    assert d['beta-delta'] == 3
    assert d['delta-gamma'] == 4
    assert d['epsilon-phi'] == 5
    assert d['phi-eta'] == 6
    assert d['phi-pi'] == 6
    assert d['phi-zeta'] == 7
    with pytest.raises(KeyError):
        d['phi-eat']  # Which would not be the case if we used 'phi-[eta|pi]' (square brackets)
    with pytest.raises(KeyError):
        d['phi-psi']
    assert d['phi-zeta-something'] == 7  # To be aware of -
    with pytest.raises(KeyError):
        d['phi-omega-something']

def test_default_handling():

    # Option 1: The usual dict method
    d = ReDict({
        'alpha-.*': 1,
        'alpha_.*': 2,
        'beta-.*': 3,
        '.*-gamma': 4,
        'eps.*': 5
        })

    assert d.get('alpha-beta', 6) == 1
    assert d.get('delta-alpha', 6) == 6
    with pytest.raises(MultipleMatchesError):
        d.get('alpha-gamma', 6)  # Multiple Matches

    # Option 2 - This is useful in ReCurseDicts where we want multiple levels of defaults.
    d = ReDict({
        'alpha-.*': 1,
        'alpha_.*': 2,
        'beta-.*': 3,
        '.*-gamma': 4,
        'eps.*': 5,
        None: 6
        })

    assert d['alpha-beta'] == 1
    assert d['delta-alpha'] == 6
    with pytest.raises(MultipleMatchesError):
        d['alpha-gamma']  # Multiple Matches


def test_recurse_dict():

    d = ReCurseDict({
        'alpha-*': {
            '.*-een': 1,
            '.*-twee': 2,
            '.*-drie': 3,
            None: -1
            },
        'beta-*': 4,
        'gamma-*': {
            'forgot_prefix': 5,
            '.*remembered_prefix': 6
            },
        None: -2,
    })

    assert d['alpha-een'] == 1
    assert d['alpha-twee'] == 2
    assert d['alpha-vier'] == -1
    assert d['beta-anything'] == 4
    with pytest.raises(KeyError):
        assert d['gamma-forgot_prefix']
    assert d['gamma-remembered_prefix'] == 6
    assert d['delta-anything'] == -2


if __name__ == '__main__':
    test_recurse_dict()
    test_default_handling()
    test_redict()



from collections import OrderedDict, namedtuple
from artemis.config import get_artemis_config_value
from artemis.plotting.matplotlib_backend import BarPlot
from matplotlib.axes import Axes
from matplotlib.gridspec import SubplotSpec
from contextlib import contextmanager
import numpy as np
from matplotlib import pyplot as plt
from artemis.plotting.drawing_plots import redraw_figure
from artemis.plotting.expanding_subplots import select_subplot
from artemis.plotting.matplotlib_backend import get_plot_from_data, TextPlot, MovingPointPlot, Moving2DPointPlot, \
    MovingImagePlot, HistogramPlot, CumulativeLineHistogram
from artemis.plotting.plotting_backend import LinePlot, ImagePlot, is_server_plotting_on

__author__ = 'peter'

"""
dbplot just takes your data, and plots it.  No fuss, no muss.  No more thinking about what kind plot to use, or how to
make updating plots of changing variables.  Just dbplot it.

    dbplot(data, 'my-data')

dbplot will look at your data, and figure out which type of plot is appropriate.  If you don't like it, you can
customize it, using the plot_type argument.

dbplot makes online plotting easy.  You want to plot updates to your variable?  Just dbplot it.

    dbplot(var, 'my-var')
    dbplot(updated_var, 'my-var')

See demo_dbplot.py for some demos of what dbplot can do.
"""


def dbplot(data, name = None, plot_type = None, axis=None, plot_mode = 'live', draw_now = True, hang = False, title=None,
           fig = None, xlabel = None, ylabel = None, draw_every = None, layout=None, legend=None, grid=False,
           wait_for_display_sec=0, cornertext = None):
    """
    Plot arbitrary data and continue execution.  This program tries to figure out what type of plot to use.

    :param data: Any data.  Hopefully, we at dbplot will be able to figure out a plot for it.
    :param name: A name uniquely identifying this plot.
    :param plot_type: A specialized constructor to be used the first time when plotting.  You can also pass
        certain string to give hints as to what kind of plot you want (can resolve cases where the given data could be
        plotted in multiple ways):
        'line': Plots a line plot
        'img': An image plot
        'colour': A colour image plot
        'pic': A picture (no scale bars, axis labels, etc).
    :param axis: A string identifying which axis to plot on.  By default, it is the same as "name".  Only use this
        argument if you indend to make multiple dbplots share the same axis.
    :param plot_mode: Influences how the data should be used to choose the plot type:
        'live': Best for 'live' plots that you intend to update as new data arrives
        'static': Best for 'static' plots, that you do not intend to update
        'image': Try to represent the plot as an image
    :param draw_now: Draw the plot now (you may choose false if you're going to add another plot immediately after and
        don't want have to draw this one again.
    :param hang: Hang on the plot (wait for it to be closed before continuing)
    :param title: Title of the plot (will default to name if not included)
    :param fig: Name of the figure - use this when you want to create multiple figures.
    :param grid: Turn the grid on
    :param wait_for_display_sec: In server mode, you can choose to wait maximally wait_for_display_sec seconds before this
        call returns. In case plotting is finished earlier, the call returns earlier. Setting wait_for_display_sec to a negative number will cause the call to block until the plot has been displayed.
    """
    if is_server_plotting_on():
        # Redirect the function call to the plotting server.  The flag gets turned on in a configuration file.  It is
        # turned off when this file is run ON the plotting server, from the first line in plotting_server.py
        arg_locals = locals().copy()
        from artemis.remote.plotting.plotting_client import dbplot_remotetly
        dbplot_remotetly(arg_locals=arg_locals)
        return

    if isinstance(fig, plt.Figure):
        assert None not in _DBPLOT_FIGURES, "If you pass a figure, you can only do it on the first call to dbplot (for now)"
        _DBPLOT_FIGURES[None] = fig
        fig = None
    elif fig not in _DBPLOT_FIGURES or not plt.fignum_exists(_DBPLOT_FIGURES[fig].figure.number):  # Second condition handles closed figures.
        _DBPLOT_FIGURES[fig] = _PlotWindow(figure = _make_dbplot_figure(), subplots=OrderedDict(), axes = {})
        if fig is not None:
            _DBPLOT_FIGURES[fig].figure.canvas.set_window_title(fig)

    suplot_dict = _DBPLOT_FIGURES[fig].subplots

    if axis is None:
        axis=name

    if name not in suplot_dict:

        if isinstance(plot_type, str):
            plot = {
                'line': LinePlot,
                'thick-line': lambda: LinePlot(plot_kwargs={'linewidth': 3}),
                'pos_line': lambda: LinePlot(y_bounds=(0, None), y_bound_extend=(0, 0.05)),
                # 'pos_line': lambda: LinePlot(y_bounds=(0, None)),
                'bar': BarPlot,
                'img': ImagePlot,
                'colour': lambda: ImagePlot(is_colour_data=True),
                'equal_aspect': lambda: ImagePlot(aspect='equal'),
                'image_history': lambda: MovingImagePlot(),
                'fixed_line_history': lambda: MovingPointPlot(buffer_len=100),
                'pic': lambda: ImagePlot(show_clims=False, aspect='equal'),
                'notice': lambda: TextPlot(max_history=1, horizontal_alignment='center', vertical_alignment='center', size='x-large'),
                'cost': lambda: MovingPointPlot(y_bounds=(0, None), y_bound_extend=(0, 0.05)),
                'percent': lambda: MovingPointPlot(y_bounds=(0, 100)),
                'trajectory': lambda: Moving2DPointPlot(axes_update_mode='expand'),
                'trajectory+': lambda: Moving2DPointPlot(axes_update_mode='expand', x_bounds=(0, None), y_bounds=(0, None)),
                'histogram': lambda: HistogramPlot(edges = np.linspace(-5, 5, 20), ),
                'cumhist': lambda: CumulativeLineHistogram(edges = np.linspace(-5, 5, 20)),
                }[plot_type]()
        elif plot_type is None:
            plot = get_plot_from_data(data, mode=plot_mode)
        else:
            assert hasattr(plot_type, "__call__")
            plot = plot_type()

        if isinstance(axis, SubplotSpec):
            axis = plt.subplot(axis)
        if isinstance(axis, Axes):
            ax = axis
            ax_name = str(axis)
        elif isinstance(axis, basestring) or axis is None:
            ax = select_subplot(axis, fig=_DBPLOT_FIGURES[fig].figure, layout=_default_layout if layout is None else layout)
            ax_name = axis
            # ax.set_title(axis)
        else:
            raise Exception("Axis specifier must be a string, an Axis object, or a SubplotSpec object.  Not {}".format(axis))

        if ax_name not in _DBPLOT_FIGURES[fig].axes:
            ax.set_title(name)
            _DBPLOT_FIGURES[fig].subplots[name] = _Subplot(axis=ax, plot_object=plot)
            _DBPLOT_FIGURES[fig].axes[ax_name] = ax

        _DBPLOT_FIGURES[fig].subplots[name] = _Subplot(axis=_DBPLOT_FIGURES[fig].axes[ax_name], plot_object=plot)
        plt.sca(_DBPLOT_FIGURES[fig].axes[ax_name])
        if xlabel is not None:
            _DBPLOT_FIGURES[fig].subplots[name].axis.set_xlabel(xlabel)
        if ylabel is not None:
            _DBPLOT_FIGURES[fig].subplots[name].axis.set_ylabel(ylabel)
        if draw_every is not None:
            _draw_counters[fig, name] = -1

        if grid:
            plt.grid()

    # Update the relevant data and plot it.  TODO: Add option for plotting update interval
    plot = _DBPLOT_FIGURES[fig].subplots[name].plot_object
    plot.update(data)
    plot.plot()

    if cornertext is not None:
        if not hasattr(_DBPLOT_FIGURES[fig].figure, '__cornertext'):
            _DBPLOT_FIGURES[fig].figure.__cornertext = _DBPLOT_FIGURES[fig].subplots.values()[0].axis.annotate(cornertext, xy=(0, 0), xytext=(0.01, 0.98), textcoords='figure fraction')
        else:
            _DBPLOT_FIGURES[fig].figure.__cornertext.set_text(cornertext)
    if title is not None:
        _DBPLOT_FIGURES[fig].subplots[name].axis.set_title(title)
    if legend is not None:
        _DBPLOT_FIGURES[fig].subplots[name].axis.legend(legend, loc='best', framealpha=0.5)

    if draw_now and not _hold_plots:
        if draw_every is not None:
            _draw_counters[fig, name]+=1
            if _draw_counters[fig, name] % draw_every != 0:
                return _DBPLOT_FIGURES[fig].subplots[name].axis
        if hang:
            plt.figure(_DBPLOT_FIGURES[fig].figure.number)
            plt.show()
        else:
            redraw_figure(_DBPLOT_FIGURES[fig].figure)
    return _DBPLOT_FIGURES[fig].subplots[name].axis


_PlotWindow = namedtuple('PlotWindow', ['figure', 'subplots', 'axes'])

_Subplot = namedtuple('Subplot', ['axis', 'plot_object'])

_DBPLOT_FIGURES = {}  # An dict<figure_name: _PlotWindow(figure, OrderedDict<subplot_name:_Subplot>)>

_DEFAULT_SIZE = get_artemis_config_value(section='plotting', option='default_fig_size', default_generator=lambda: (10, 8), write_default=True, read_method='eval')

_draw_counters = {}

_hold_plots = False

_hold_plot_counter = 0

_default_layout = 'grid'


def reset_dbplot():
    for fig_name, plot_window in _DBPLOT_FIGURES.items():
        plt.close(plot_window.figure)
        del _DBPLOT_FIGURES[fig_name]


def set_dbplot_figure_size(width, height):
    global _DEFAULT_SIZE
    _DEFAULT_SIZE = (width, height)


def set_dbplot_default_layout(layout):
    global _default_layout
    _default_layout = layout


def get_dbplot_figure(name=None):
    return _DBPLOT_FIGURES[name].figure


def get_dbplot_subplot(name, fig_name=None):
    return _DBPLOT_FIGURES[fig_name].subplots[name].axis


def _make_dbplot_figure():

    if _DEFAULT_SIZE is None:
        fig= plt.figure()
    else:
        fig= plt.figure(figsize=_DEFAULT_SIZE)  # This is broken in matplotlib2 for some reason

    # fig.cornerbox__ = fig.add_axes([0, 0, 0.2, 0.05])
    return fig


def freeze_dbplot(name, fig = None):
    del _DBPLOT_FIGURES[fig].subplots[name]


def freeze_all_dbplots(fig = None):
    for name in _DBPLOT_FIGURES[fig].subplots.keys():
        freeze_dbplot(name, fig=fig)


@contextmanager
def hold_dbplots(fig = None, draw_every = None):
    """
    Use this in a "with" statement to prevent plotting until the end.
    :param fig:
    :return:
    """
    if is_server_plotting_on():
        # For now, this does nothing.  Eventually, it should be made to send a "draw" command through the pipe
        yield
        return

    global _hold_plots
    _old_hold_state = _hold_plots
    _hold_plots = True
    yield
    _hold_plots = _old_hold_state

    if _old_hold_state:
        plot_now = False
    elif draw_every is not None:
        global _hold_plot_counter
        plot_now = _hold_plot_counter % draw_every == 0
        _hold_plot_counter+=1
    else:
        plot_now = True

    if plot_now and fig in _DBPLOT_FIGURES:
        redraw_figure(_DBPLOT_FIGURES[fig].figure)


def clear_dbplot(fig = None):
    if fig in _DBPLOT_FIGURES:
        plt.figure(_DBPLOT_FIGURES[fig].figure.number)
        plt.clf()
        _DBPLOT_FIGURES[fig].subplots.clear()
        _DBPLOT_FIGURES[fig].axes.clear()


def get_dbplot_axis(axis_name, fig=None):
    """
    Get the named axis of a dbplot.
    """
    return _DBPLOT_FIGURES[fig].axes[axis_name]


def dbplot_hang():
    plt.show()


def dbplot_collection(collection, name, **kwargs):
    """
    Plot a collection of items in one go.
    :param collection:
    :param name:
    :param kwargs:
    :return:
    """
    with hold_dbplots():
        if isinstance(collection, (list, tuple)):
            for i, el in enumerate(collection):
                dbplot(el, '{}[{}]'.format(name, i), **kwargs)



from numpy.random.mtrand import RandomState
import numpy as np
__author__ = 'peter'


def get_rng(initializer):
    """
    Creates a random number generator from either a seed, the time, or another random number generator.
    This is useful when you need to initialize random states on many levels, and you want the choice between
    specifying a seed directly or passing in an RNG from somewhere else.

    :param initializer: Can be:
        None, in which case the RNG is seeded on the time
        an int, in which case this is used as the seed
        a numpy RandomState object, in which case it's just passed through.
    :return: A numpy RandomState object.
    """
    if initializer is None or isinstance(initializer, int):
        return np.random.RandomState(initializer)
    elif isinstance(initializer, RandomState):
        return initializer



import matplotlib.pyplot as plt
import logging
import numpy as np
import matplotlib.colors as colors
from si_prefix import si_format

logging.basicConfig()
ARTEMIS_LOGGER = logging.getLogger('artemis')

"""
An few extension functions to pyplot
"""


def axhlines(ys, **plot_kwargs):
    """
    Draw horizontal lines across plot
    :param ys: A scalar, list, or 1D array of vertical offsets
    :param plot_kwargs: Keyword arguments to be passed to plot
    :return: The plot object corresponding to the lines.
    """
    ys = np.array((ys, ) if np.isscalar(ys) else ys, copy=False)
    lims = plt.gca().get_xlim()
    y_points = np.repeat(ys[:, None], repeats=3, axis=1).flatten()
    x_points = np.repeat(np.array(lims + (np.nan, ))[None, :], repeats=len(ys), axis=0).flatten()
    plot = plt.plot(x_points, y_points, scalex = False, **plot_kwargs)
    return plot


def axvlines(xs, **plot_kwargs):
    """
    Draw vertical lines on plot
    :param xs: A scalar, list, or 1D array of horizontal offsets
    :param plot_kwargs: Keyword arguments to be passed to plot
    :return: The plot object corresponding to the lines.
    """
    xs = np.array((xs, ) if np.isscalar(xs) else xs, copy=False)
    lims = plt.gca().get_ylim()
    x_points = np.repeat(xs[:, None], repeats=3, axis=1).flatten()
    y_points = np.repeat(np.array(lims + (np.nan, ))[None, :], repeats=len(xs), axis=0).flatten()
    plot = plt.plot(x_points, y_points, scaley = False, **plot_kwargs)
    return plot

    # return [plt.axhline(y, **specs) for y in ys]
    # return [plt.axvline(x, **specs) for x in xs]


def set_default_figure_size(width, height):
    """
    :param width: Width (in inches, for some reason)
    :param height: Height (also in inches.  One inch is about 2.54cm)
    """
    from pylab import rcParams
    rcParams['figure.figsize'] = width, height


_lines_colour_cycle = [p['color'] for p in plt.rcParams['axes.prop_cycle']]


def get_lines_color_cycle():
    return _lines_colour_cycle


def get_line_color(ix, modifier=None):
    colour = _lines_colour_cycle[ix]
    if modifier=='dark':
        return tuple(c/2 for c in colors.hex2color(colour))
    elif modifier=='light':
        return tuple(1-(1-c)/2 for c in colors.hex2color(colour))
    elif modifier is not None:
        raise NotImplementedError(modifier)
    return colors.hex2color(colour)


def non_uniform_imshow(im, x_locs=None, y_locs=None, spacing='lin', format_str='{:.2g}', **other_imagesc_args):

    if x_locs is not None:
        assert len(x_locs)==im.shape[1]
        assert np.all(np.diff(x_locs)>0)
    if y_locs is not None:
        assert len(y_locs)==im.shape[0]
        assert np.all(np.diff(y_locs)>0)

    handle = plt.imshow(im, **other_imagesc_args)
    plt.gca().invert_yaxis()

    relabel_axes(plt.gca(), xvalues=x_locs, yvalues=y_locs, format_str=format_str)
    return handle


def relabel_axis(axis, value_array, n_points = 5, format_str='{:.2g}'):
    ticks = np.round(np.linspace(0, len(value_array)-1, num=n_points)).astype('int')
    axis.set_ticks(ticks)
    if format_str=='SI':
        axis.set_ticklabels([si_format(t, format_str='{value}{prefix}') for t in value_array[ticks]])
    else:
        axis.set_ticklabels([format_str.format(t) for t in value_array[ticks]])


def relabel_axes(axes, n_points='auto', xvalues=None, yvalues=None, xlabel=None, ylabel=None, format_str='{:.2g}'):

    if n_points=='auto':
        n_points = 5 if len(xvalues)%5==0 else \
            4 if len(xvalues)%4==0 else \
            3 if len(xvalues)%3==0 else \
            5
    if xvalues is not None:
        relabel_axis(axes.xaxis, xvalues, n_points=n_points, format_str=format_str)
    if xlabel is not None:
        axes.set_xlabel(xlabel)
    if yvalues is not None:
        relabel_axis(axes.yaxis, yvalues, n_points=n_points, format_str=format_str)
    if ylabel is not None:
        axes.set_ylabel(ylabel)


def remove_x_axis():
    plt.tick_params(axis='x', labelbottom='off')


def remove_y_axis():
    plt.tick_params(axis='y', labelbottom='off')
    # plt.gca().yaxis.set_major_locator(plt.NullLocator())



import numpy as np
from artemis.plotting.db_plotting import dbplot
from artemis.plotting.bokeh_backend import set_url, LinePlot, MovingPointPlot, HistoryPlot
# import os
# os.environ["DISPLAY"] = "localhost:10.0"
import time

def tet_dbplot(n_steps = 3):

    arr = np.random.rand(10, 10)

    for i in xrange(n_steps):
        arr_sq=arr**2
        arr = arr_sq/np.mean(arr_sq)
        barr = np.array([[x**10 for x in np.arange(0,5)], [x**10 for x in np.arange(0,5)]])
        barr = barr.T
        barr2 = np.array([[x**9 for x in np.arange(0,5)], [x**8 for x in np.arange(0,5)]])
        barr2 = barr2.T
        kw = {"y_axis_type":"log"}
        dbplot(barr, 'barr', plot_type=lambda: LinePlot(**kw))
        dbplot(barr2, 'barr2', plot_type=lambda: LinePlot(**kw))


def moving_point_plot(n_steps = 20):
    for i in xrange(n_steps):
        data = np.array([i,i**2])
        if i == 4:
            data = np.array([float('nan'), i**2])
        dbplot(data, "history", plot_type=lambda: MovingPointPlot())
        time.sleep(0.5)

if __name__ == '__main__':
    # test_particular_plot()
    set_url("http://146.50.149.168:5006")
    moving_point_plot(n_steps = 20)
    time.sleep(2)
    tet_dbplot()




import pytest
from artemis.plotting.plotting_backend import get_plotting_server_address
from artemis.remote.utils import get_local_ips
from artemis.remote.virtualenv import check_diff_local_remote_virtualenv, get_remote_installed_packages


ip_address = get_plotting_server_address()
is_local = ip_address in get_local_ips()

@pytest.mark.skipif(ip_address="" or is_local, reason ="No sense for local ip")
def test_check_diff_local_remote_virtualenv():

    # original_virtual_env_value = get_config_value(".artemisrc", ip_address, "python")
    # import ConfigParser
    # import os
    #
    # Config = ConfigParser.ConfigParser()
    # Config.read(os.path.expanduser("~/.artemisrc"))
    # Config.set(section=ip_address,option="python",value="~/virtualenvs/test_env/bin/python")
    # with open(os.path.expanduser("~/.artemisrc"), 'wb') as configfile:
    #     Config.write(configfile)

    check_diff_local_remote_virtualenv(ip_address, auto_install=False, auto_upgrade=False,ignore_warnings=True)

    # Config.set(section=ip_address,option="python",value=original_virtual_env_value)
    # with open(os.path.expanduser("~/.artemisrc"), 'wb') as configfile:
    #     Config.write(configfile)

@pytest.mark.skipif(ip_address="" or is_local, reason ="No sense for local ip")
def test_get_remote_installed_packages():
    packages = get_remote_installed_packages(ip_address)
    print packages


if __name__ == "__main__":
    test_get_remote_installed_packages()
    test_check_diff_local_remote_virtualenv()


from collections import OrderedDict
from functools import partial

from artemis.experiments.experiment_record import record_experiment
from artemis.experiments.experiments import Experiment, GLOBAL_EXPERIMENT_LIBRARY


# ALTERNATE INTERFACES.
# We keep these for backwards compatibility and to show how else we could organize the experiment API.
# These are out of use, as we at Artemis inc. prefer the @experiment_function decorator.


# Register interface:

def register_experiment(name, function, description = None, conclusion = None, versions = None, current_version = None, **kwargs):
    """
    This is the old interface to experiments.  We keep it, for now, for the sake of
    backwards-compatibility.

    In the future, use the @experiment_function decorator instead.
    """
    info = OrderedDict()
    if description is not None:
        info['Description'] = description
    if conclusion is not None:
        info['Conclusion'] = conclusion

    if versions is not None:
        if current_version is None:
            assert len(versions)==1
            current_version = versions.keys()[0]
        assert current_version is not None
        function = partial(function, **versions[current_version])

    assert name not in GLOBAL_EXPERIMENT_LIBRARY, 'An experiment with name "%s" has already been registered!' % (name, )
    experiment = Experiment(name = name, function=function, info=info, **kwargs)
    GLOBAL_EXPERIMENT_LIBRARY[name] = experiment
    return experiment


# Start/Stop Interface:

_CURRENT_EXPERIMENT_CONTEXT = None


def start_experiment(*args, **kwargs):
    global _CURRENT_EXPERIMENT_CONTEXT
    _CURRENT_EXPERIMENT_CONTEXT = record_experiment(*args, **kwargs)
    return _CURRENT_EXPERIMENT_CONTEXT.__enter__()


def end_current_experiment():
    global _CURRENT_EXPERIMENT_CONTEXT
    _CURRENT_EXPERIMENT_CONTEXT.__exit__(None, None, None)
    _CURRENT_EXPERIMENT_CONTEXT = None


# Experiment Library interface:


class _ExpLibClass(object):

    def __setattr__(self, experiment_name, experiment):
        assert isinstance(experiment, Experiment), "Your experiment must be an experiment!"
        if experiment.name is None:
            experiment.name = experiment_name
        assert experiment_name not in GLOBAL_EXPERIMENT_LIBRARY, "Experiment %s is already in the library" % (experiment_name, )
        self.__dict__[experiment_name] = experiment
        GLOBAL_EXPERIMENT_LIBRARY[experiment_name] = experiment

    def get_experiments(self):
        return GLOBAL_EXPERIMENT_LIBRARY

    def __getattr__(self, name):
        if name in GLOBAL_EXPERIMENT_LIBRARY:
            return GLOBAL_EXPERIMENT_LIBRARY[name]
        else:
            return _ExperimentConstructor(name)


class _ExperimentConstructor(object):

    def __init__(self, name):
        self.name = name

    def __call__(self, **kwargs):
        if self.name in GLOBAL_EXPERIMENT_LIBRARY:
            raise Exception("You tried to run create experiment '%s', but it already exists in the library.  Give it another name!" % (self.name, ))
        return register_experiment(name = self.name, **kwargs)

    def run(self):
        raise Exception("You tried to run experiment '%s', but it hasn't been made yet!" % (self.name, ))


ExperimentLibrary = _ExpLibClass()





import numpy as np

__author__ = 'peter'


def array_info(arr):
    """
    :param arr: Return info about an array
    :return:
    """
    return '<{dtype} array of shape {shape}: mean: {mean:.3g}, std: {std:.3g}, min: {min:.3g}, max: {max:.3g} at {addr:s}>'.format(
        dtype = arr.dtype,
        shape = arr.shape,
        mean = arr.mean(),
        std = arr.std(),
        min = arr.min(),
        max = arr.max(),
        addr = hex(id(arr))
        )


def argtopk(x, k, axis=-1):
    return np.take(np.argpartition(-x, axis=axis, kth=k), np.arange(k), axis=axis)





import shutil
from artemis.fileman.file_getter import get_file_in_archive
from artemis.fileman.local_dir import get_local_path
import os
__author__ = 'peter'


def test_unpack_tar_gz():

    if os.path.exists(get_local_path('tests/test_tar_zip')):
        shutil.rmtree(get_local_path('tests/test_tar_zip'))
    if os.path.exists(get_local_path('tests/test_tar_zip.tar.gz')):
        os.remove(get_local_path('tests/test_tar_zip.tar.gz'))

    for _ in xrange(2):  # (Second time to check caching)

        local_file = get_file_in_archive(
            relative_path= 'tests/test_tar_zip',
            url = 'https://drive.google.com/uc?export=download&id=0B4IfiNtPKeSAbmp6VEVJdjdSSlE',
            subpath = 'testzip/test_file.txt'
            )

        with open(local_file) as f:
            txt = f.read()

        assert txt == 'blah blah blah'


def test_unpack_zip():

    if os.path.exists(get_local_path('tests/test_zip_zip')):
        shutil.rmtree(get_local_path('tests/test_zip_zip'))
    if os.path.exists(get_local_path('tests/test_zip_zip.zip')):
        os.remove(get_local_path('tests/test_zip_zip.zip'))

    for _ in xrange(2):  # (Second time to check caching)

        local_file = get_file_in_archive(
            relative_path= 'tests/test_zip_zip',
            url = 'https://drive.google.com/uc?export=download&id=0B4IfiNtPKeSATWZXWjEyd1FsRG8',
            subpath = 'testzip/test_file.txt'
            )

        with open(local_file) as f:
            txt = f.read()

        assert txt == 'blah blah blah'


if __name__ == '__main__':
    test_unpack_zip()
    test_unpack_tar_gz()




import getpass
import multiprocessing
import traceback
from collections import OrderedDict
from functools import partial
from importlib import import_module

from artemis.experiments.experiment_record import load_experiment_record, ExpInfoFields, \
    ExpStatusOptions, ARTEMIS_LOGGER, record_id_to_experiment_id
from artemis.experiments.experiments import load_experiment, GLOBAL_EXPERIMENT_LIBRARY
from artemis.fileman.config_files import get_home_dir
from artemis.general.hashing import compute_fixed_hash
from artemis.general.should_be_builtins import izip_equal, detect_duplicates


def pull_experiments(user, ip, experiment_names, include_variants=True):
    """
    Pull experiments from another computer matching the given experiment name.

    :param user:
    :param ip:
    :param experiment_name:
    :param include_variants:
    :return:
    """
    import pexpect
    import sys

    if isinstance(experiment_names, basestring):
        experiment_names = [experiment_names]

    inclusions = ' '.join("--include='**/*-{exp_name}{variants}/*'".format(exp_name=exp_name, variants = '*' if include_variants else '') for exp_name in experiment_names)

    home = get_home_dir()

    command = "rsync -a -m -i {inclusions} --include='*/' --exclude='*' {user}@{ip}:~/.artemis/experiments/ {home}/.artemis/experiments/".format(
        inclusions=inclusions,
        user=user,
        ip=ip,
        home=home
        )
    password = getpass.getpass("Enter password for {}@{}:".format(user, ip))
    child = pexpect.spawn(command)
    code = child.expect([pexpect.TIMEOUT, 'password:'])
    if code == 0:
        print("Got unexpected output: %s %s" % (child.before, child.after))
        sys.exit()
    else:
        child.sendline(password)
    output = child.read()
    return output


def load_lastest_experiment_results(experiments, error_if_no_result = True):
    """
    :param experiments:
    :param error_if_no_result:
    :return:
    """
    results = OrderedDict()
    for ex in experiments:
        record = ex.get_latest_record(err_if_none=error_if_no_result, only_completed=True)
        if record is None:
            if error_if_no_result:
                raise Exception("Experiment {} had no result.  Run this experiment to completion before trying to compare its results.".format(ex.get_id()))
            else:
                ARTEMIS_LOGGER.warn('Experiment {} had no records.  Not including this in results'.format(ex.get_id()))
        else:
            results[ex.get_id()] = record.get_result()
    if len(results)==0:
        ARTEMIS_LOGGER.warn('None of your experiments had any results.  Your comparison function will probably show no meaningful result.')
    return results


def select_experiments(user_range, exp_record_dict, return_dict=False):

    exp_filter = _filter_experiments(user_range, exp_record_dict)
    if return_dict:
        return OrderedDict((name, exp_record_dict[name]) for name in exp_record_dict if exp_filter[name])
    else:
        return [name for name in exp_record_dict if exp_filter[name]]


def _filter_experiments(user_range, exp_record_dict):

    if user_range in exp_record_dict:
        is_in = [k==user_range for k in exp_record_dict]
    else:
        number_range = interpret_numbers(user_range)
        if number_range is not None:
            # experiment_ids = [experiment_list[i] for i in number_range]
            is_in = [i in number_range for i in xrange(len(exp_record_dict))]
        elif user_range == 'all':
            # experiment_ids = experiment_list
            is_in = [True]*len(exp_record_dict)
        elif user_range.startswith('has:'):
            phrase = user_range[len('has:'):]
            # experiment_ids = [exp_id for exp_id in experiment_list if phrase in exp_id]
            is_in = [phrase in exp_id for exp_id in exp_record_dict]
        elif user_range.startswith('1diff:'):
            # select experiments whose arguments differ by one element from the selected experiments
            base_range = user_range[len('1diff:'):]
            base_range_exps = select_experiments(base_range, exp_record_dict) # list<experiment_id>
            all_exp_args_hashes = {eid: set(compute_fixed_hash(a) for a in load_experiment(eid).get_args().items()) for eid in exp_record_dict} # dict<experiment_id : set<arg_hashes>>
            # assert all_equal_length(all_exp_args_hashes.values()), 'All variants must have the same number of arguments' # Note: we diable this because we may have lists of experiments with different root functions.
            is_in = [any(len(all_exp_args_hashes[eid].difference(all_exp_args_hashes[other_eid]))<=1 for other_eid in base_range_exps) for eid in exp_record_dict]
        elif user_range.startswith('hasnot:'):
            phrase = user_range[len('hasnot:'):]
            # experiment_ids = [exp_id for exp_id in experiment_list if phrase not in exp_id]
            is_in = [phrase not in exp_id for exp_id in exp_record_dict]
        elif user_range in ('unfinished', 'invalid'):  # Return all experiments where all records are unfinished/invalid
            record_filters = _filter_records(user_range, exp_record_dict)
            # experiment_ids = [exp_id for exp_id in experiment_list if len(record_filters[exp_id])]
            is_in = [all(record_filters[exp_id]) for exp_id in exp_record_dict]
        else:
            raise Exception("Don't know how to use input '{}' to select experiments".format(user_range))

    return OrderedDict((exp_id, exp_is_in) for exp_id, exp_is_in in izip_equal(exp_record_dict, is_in))


def select_experiment_records(user_range, exp_record_dict, flat=True):
    """
    :param user_range:
    :param exp_record_dict: An OrderedDict<experiment_name: list<experiment_record_name>>
    :param flat: Return a list of experiment records, instead of an OrderedDict
    :return: if not flat, an An OrderedDict<experiment_name: list<experiment_record_name>>
        otherwise a list<experiment_record_name>
    """
    filters = _filter_records(user_range, exp_record_dict)
    filtered_dict = OrderedDict((k, [load_experiment_record(record_id) for record_id, f in izip_equal(exp_record_dict[k], filters[k]) if f]) for k in exp_record_dict.keys())
    if flat:
        return [record_id for records in filtered_dict.values() for record_id in records]
    else:
        return filtered_dict


def _filter_records(user_range, exp_record_dict):
    """
    :param user_range:
    :param exp_record_dict:
    :return: An OrderedDict<experiment_id -> list<True or False>> indicating whether each record from the given experiment passed the filter
    """

    def _bitwise(op, filter_set_1, filter_set_2):
        assert op in ('and', 'or')
        filter_set_3 = filter_set_1.copy()
        for k in filter_set_1.keys():
            filter_set_3[k] = [(a or b) if op=='or' else (a and b) for a, b in izip_equal(filter_set_1[k], filter_set_2[k])]
        return filter_set_3

    base = OrderedDict((k, [False]*len(v)) for k, v in exp_record_dict.iteritems())
    if user_range in exp_record_dict:  # User just lists an experiment
        base[user_range] = [True]*len(base[user_range])
        return base

    if '|' in user_range:
        return reduce(lambda a, b: _bitwise('or', a, b), [_filter_records(subrange, exp_record_dict) for subrange in user_range.split('|')])
    if '&' in user_range:
        return reduce(lambda a, b: _bitwise('and', a, b), [_filter_records(subrange, exp_record_dict) for subrange in user_range.split('&')])
    number_range = interpret_numbers(user_range)
    keys = exp_record_dict.keys()
    if number_range is not None:
        for i in number_range:
            base[keys[i]] = [True]*len(base[keys[i]])
    elif '.' in user_range:
        exp_rec_pairs = interpret_record_identifier(user_range)
        for exp_number, rec_number in exp_rec_pairs:
            base[keys[exp_number]][rec_number] = True
    elif user_range == 'old':
        for k, v in base.iteritems():
            base[k] = ([True]*(len(v)-1)+[False]) if len(v)>0 else []
    elif user_range == 'unfinished':
        for k, v in base.iteritems():
            base[k] = [load_experiment_record(rec_id).info.get_field(ExpInfoFields.STATUS) != ExpStatusOptions.FINISHED for rec_id in exp_record_dict[k]]
        # filtered_dict = OrderedDict((exp_id, [rec_id for rec_id in records if load_experiment_record(rec_id).info.get_field(ExpInfoFields.STATUS) != ExpStatusOptions.FINISHED]) for exp_id, records in exp_record_dict.iteritems())
    elif user_range == 'invalid':
        for k, v in base.iteritems():
            base[k] = [load_experiment_record(rec_id).is_valid() is False for rec_id in exp_record_dict[k]]
    elif user_range == 'all':
        for k, v in base.iteritems():
            base[k] = [True]*len(v)
    elif user_range == 'errors':
        for k, v in base.iteritems():
            base[k] = [load_experiment_record(rec_id).info.get_field(ExpInfoFields.STATUS)==ExpStatusOptions.ERROR for rec_id in exp_record_dict[k]]
    else:
        raise Exception("Don't know how to interpret subset '{}'".format(user_range))
    return base


def _filter_experiment_record_list(user_range, experiment_record_ids):
    if user_range=='all':
        return [True]*len(experiment_record_ids)
    elif user_range=='new':
        return detect_duplicates(experiment_record_ids, key=record_id_to_experiment_id, keep_last=True)
        # return [n for n, is_old in izip_equal(get_record_ids(), old) if not old]
    elif user_range=='old':
        return [not x for x in _filter_records(user_range, 'new')]
    elif user_range=='orphans':
        orphans = []
        for i, record_id in enumerate(experiment_record_ids):
            info = load_experiment_record(record_id).info
            if 'Module' in info:
                try:
                    import_module(info['Module'])
                    if not record_id_to_experiment_id(record_id) in GLOBAL_EXPERIMENT_LIBRARY:
                        orphans.append(True)
                    else:
                        orphans.append(False)
                except ImportError:
                    orphans.append(True)
            else:  # They must be old... lets kill them!
                orphans.append(True)
        return orphans
    else:
        which_ones = interpret_numbers(user_range)
        if which_ones is None:
            raise Exception('Could not interpret user range: "{}"'.format(user_range))
        filters = [False]*len(experiment_record_ids)
        for i in which_ones:
            filters[i] = True
        return filters


def select_experiment_records_from_list(user_range, experiment_records):
    return [rec_id for rec_id, f in izip_equal(experiment_records, _filter_experiment_record_list(user_range, experiment_records)) if f]


def interpret_record_identifier(user_range):
    """
    You can identify a single record with, eg 3.4, meaning "record 4 from experiment 3:.
    You can identify a range with, eg 3.1-3, meaning "records 3.1, 3.2, 3.3"
    :param user_range: The user input
    :return: A list of 2 tuples (exp_no, record_no).  e.g [(3, 4)], or [(3, 1), (3, 2), (3, 3)] in the above examples.
    """
    if ',' in user_range:
        parts = user_range.split(',')
        return [pair for p in parts for pair in interpret_record_identifier(p)]
    if '.' not in user_range:
        return None
    else:
        exp_number, record_numbers = user_range.split('.')
        return [(int(exp_number), rec_num) for rec_num in interpret_numbers(record_numbers)]


def interpret_numbers(user_range):
    """
    :param user_range: A string specifying a range of numbers.  Eg.
        interpret_numbers('4-6')==[4,5,6]
        interpret_numbers('4,6')==[4,6]
        interpret_numbers('4,6-9')==[4,6,7,8,9]
    :return: A list of integers, or None if the input is not numberic
    """
    if all(d in '0123456789-,' for d in user_range):
        numbers_and_ranges = user_range.split(',')
        numbers = [n for lst in [[int(s)] if '-' not in s else range(int(s[:s.index('-')]), int(s[s.index('-')+1:])+1) for s in numbers_and_ranges] for n in lst]
        return numbers
    else:
        return None


def run_experiment(name, exp_dict=GLOBAL_EXPERIMENT_LIBRARY, **experiment_record_kwargs):
    """
    Run an experiment and save the results.  Return a string which uniquely identifies the experiment.
    You can run the experiment agin later by calling show_experiment(location_string):

    :param name: The name for the experiment (must reference something in exp_dict)
    :param exp_dict: A dict<str:func> where funcs is a function with no arguments that run the experiment.
    :param experiment_record_kwargs: Passed to ExperimentRecord.

    :return: A location_string, uniquely identifying the experiment.
    """
    experiment = exp_dict[name]
    return experiment.run(**experiment_record_kwargs)


def run_experiment_ignoring_errors(name, **kwargs):
    try:
        return run_experiment(name, **kwargs)
    except Exception as err:
        traceback.print_exc()


def run_multiple_experiments(experiments, parallel = False, cpu_count=None, raise_exceptions=True, run_args = {}):
    """
    Run multiple experiments, optionally in parallel with multiprocessing.

    :param experiments: A collection of experiments
    :param parallel: True to run in parallel, with multiprocessing
    :param cpu_count: If parallel, number of CPUs to use (defaults to all)
    :param raise_exceptions: Terminate exectution when one experiment fails.
    :param run_args: Other args to pass to Experiment.run()
    :return: A collection of experiment records.
    """

    if parallel:
        experiment_identifiers = [ex.get_id() for ex in experiments]
        if cpu_count is None:
            cpu_count = multiprocessing.cpu_count()
        func = run_experiment if raise_exceptions else run_experiment_ignoring_errors
        p = multiprocessing.Pool(processes=cpu_count)
        return p.map(partial(func, **run_args), experiment_identifiers)
    else:
        return [ex.run(raise_exceptions=raise_exceptions, display_results=False, **run_args) for ex in experiments]



import numpy as np

from artemis.ml.predictors.i_predictor import IPredictor


class Perceptron(IPredictor):

    def __init__(self, w, b = None, alpha = 0.1):
        self._w = w
        self._b = np.zeros((1, w.shape[1])) if b is None else b
        self._alpha = alpha

    def train(self, x, target_data):
        y = (self.predict(x) > 0).astype(int)
        self._w += self._alpha*x.T.dot(target_data-y)
        self._b += self._alpha*np.sum(target_data-y, axis = 0)

    def predict(self, x):
        return x.dot(self._w)+self._b

    def to_categorical(self, **kwargs):
        from artemis.ml.predictors.i_predictor import CategoricalPredictor
        return CategoricalPredictor(self, **kwargs)



import os
import sys
from ConfigParser import NoSectionError, NoOptionError, ConfigParser

__author__ = 'peter'


def get_config_value(config_filename, section, option, default_generator = None, write_default = False, read_method=None):
    """
    Get a setting from a configuration file.  If none exists, you can optionally create one.  An example config file is
    the ~/.theanorc file, which may contain:

        [global]                    # <-- section
        floatX = float32            # <-- option (value is 'float32')
        device = gpu0

        [lib]
        cnmem = 1

        [cuda]
        root=/usr/local/cuda/

    :param config_filename: A configuration filename (eg '.theanorc').  Here we enforce the convention that files start
        with '.', and are stored in the user-home directory.
    :param section: Section in the config file (as referred by squared brackets: see example above)
    :param option: The option of interest (see above example)
    :param default_generator: A function that generates the property if it is not there, or just the default value if
        default_generator is not callable
    :param write_default: Set to true if property was not found and you want to write the default value into the file.
    :param read_method: The method to read your setting.
        If left at None (default) it just returns the string.
        If 'eval' it parses the setting into a python object
        If it is a function, it passes the value through the function before returning it.
    :return: The value of the property of interest.
    """
    config_path = get_config_path(config_filename)
    default_used = False

    if write_default:
        assert default_generator is not None, "If you set write_default true, you must provide a function that can generate the default."

    if not os.path.exists(config_path):
        assert default_generator is not None, 'No config file "%s" exists, and you do not have any default value.' % (config_path, )
        value = default_generator()
        default_used = True

    else:
        config = ConfigParser()
        config.read(config_path)
        try:
            value = config.get(section, option)
        except (NoSectionError, NoOptionError) as err:
            if default_generator is None:
                raise
            else:
                value = default_generator() if callable(default_generator) else default_generator
                default_used = True

    if default_used and write_default:
        config = ConfigParser()
        config.read(config_path)
        if not config.has_section(section):
            config.add_section(section)
        config.set(section, option, value)
        with open(config_path, 'w') as f:
            config.write(f)

    if read_method == 'eval':
        value = eval(value)
    elif callable(read_method):
        value = read_method(value)
    return value


def _get_config_object(config_filename):
    config_path = get_config_path(config_filename)
    assert os.path.exists(config_path), '"{}" does not exist!'.format(config_path)
    config = ConfigParser()
    config.read(config_path)
    return config


def get_config_sections(config_filename):

    sections = _get_config_object(config_filename).sections()
    return sections


def get_home_dir():
    # This function exists because of some weirdness when running remotely.
    home_dir = os.path.expanduser('~')
    if os.path.exists(home_dir):
        return home_dir
    elif sys.platform=='linux2':
        home_dir = os.path.join('/home', os.getlogin())
    elif sys.platform=='darwin':
        home_dir==os.path.join('/Users', os.getlogin())
    assert os.path.exists(home_dir), 'The home directory "{}" seems not to exist.  This is odd, to say the least.'.format(home_dir)
    return home_dir


def get_config_path(config_filename):
    assert config_filename.startswith('.'), "We enforce the convention that configuration files must start with '.'"
    config_path = os.path.join(get_home_dir(), config_filename)
    return config_path


if __name__ == '__main__':
    config_path = get_config_path('.artemisrc')
    print 'Contents of {}:\n-------------\n'.format(config_path)
    with open(config_path) as f:
        print f.read()



import numpy as np
from artemis.general.ezprofile import EZProfiler
from artemis.general.pareto_efficiency import is_pareto_efficient_ixs, is_pareto_efficient_dumb, \
    is_pareto_efficient

__author__ = 'peter'


def test_is_pareto_efficient(plot=False):

    for n_costs in (2, 10):

        rng = np.random.RandomState(1234)

        costs = rng.rand(1000, n_costs)
        ixs = is_pareto_efficient_ixs(costs)

        assert np.sum(ixs)>0
        for c in costs[ixs]:
            assert np.all(np.any(c<=costs, axis=1))

        if plot and n_costs==2:
            import matplotlib.pyplot as plt
            plt.plot(costs[:, 0], costs[:, 1], '.')
            plt.plot(costs[ixs, 0], costs[ixs, 1], 'ro')
            plt.show()


def profile_pareto_efficient():

    rng = np.random.RandomState(1234)

    costs = rng.rand(5000, 2)

    with EZProfiler('dumb'):
        dumb_ixs = is_pareto_efficient_dumb(costs)

    with EZProfiler('smart'):
        less_dumb__ixs = is_pareto_efficient(costs)
    assert np.array_equal(dumb_ixs, less_dumb__ixs)

    with EZProfiler('index-tracking'):
        smart_ixs = is_pareto_efficient_ixs(costs)

    assert np.array_equal(dumb_ixs, smart_ixs)


if __name__ == '__main__':
    test_is_pareto_efficient()



from artemis.general.should_be_builtins import itermap, reducemap, separate_common_items, remove_duplicates, \
    detect_duplicates, remove_common_prefix, all_equal

__author__ = 'peter'


def test_reducemap():
    assert reducemap(lambda a, b: a+b, [1, 2, -4, 3, 6, -7], initial=0) == [1, 3, -1, 2, 8, 1]
    assert reducemap(lambda a, b: a+b, [1, 2, -4, 3, 6, -7]) == [3, -1, 2, 8, 1]
    assert reducemap(lambda a, b: a+b, [1, 2, -4, 3, 6, -7], include_zeroth=True) == [1, 3, -1, 2, 8, 1]


def test_itermap():
    # See collatz conjecture
    assert itermap(lambda a: a/2 if a % 2==0 else a*3+1, initial = 12, stop_func=lambda x: x==1, include_zeroth=True) == [12, 6, 3, 10, 5, 16, 8, 4, 2, 1]
    assert itermap(lambda a: a/2 if a % 2==0 else a*3+1, initial = 1, n_steps=5, include_zeroth=True) == [1, 4, 2, 1, 4, 2]


def test_separate_common_items():

    x={'a': 1, 'b':2, 'c':3}
    y={'a': 1, 'b':4, 'c':8}
    z={'a': 1, 'b':2, 'c':9}
    common, separate = separate_common_items([x, y, z])
    assert common == {'a': 1}
    assert separate == [{'b':2, 'c':3}, {'b':4, 'c':8}, {'b':2, 'c':9}]


def test_remove_duplicates():
    assert remove_duplicates(['a', 'b', 'a', 'c', 'c'])==['a', 'b', 'c']
    assert remove_duplicates(['a', 'b', 'a', 'c', 'c'], keep_last=True)==['b', 'a', 'c']
    assert remove_duplicates(['Alfred', 'Bob', 'Cindy', 'Alina', 'Karol', 'Betty'], key=lambda x: x[0])==['Alfred', 'Bob', 'Cindy', 'Karol']
    assert remove_duplicates(['Alfred', 'Bob', 'Cindy', 'Alina', 'Karol', 'Betty'], key=lambda x: x[0], keep_last=True)==['Cindy', 'Alina', 'Karol', 'Betty']
    assert remove_duplicates(['Alfred', 'Bob', 'Cindy', 'Alina', 'Karol', 'Betty'], key=lambda x: x[0], keep_last=True, hashable=False)==['Cindy', 'Alina', 'Karol', 'Betty']


def test_detect_duplicates():
    assert detect_duplicates(['a', 'b', 'a', 'c', 'c'])==[False, False, True, False, True]
    assert detect_duplicates(['a', 'b', 'a', 'c', 'c'], keep_last=True)==[True, False, False, True, False]


def test_remove_common_prefix():
    assert remove_common_prefix([[1, 2, 3, 4], [1, 2, 5], [1, 2, 3, 5]]) == [[3, 4], [5], [3, 5]]
    assert remove_common_prefix([[1, 2, 3, 4], [1, 2, 5], [1, 2, 3, 5]], max_elements=1) == [[2, 3, 4], [2, 5], [2, 3, 5]]


def test_all_equal():

    assert all_equal([2, 2, 2])
    assert not all_equal([2, 2, 3])
    assert all_equal([])

if __name__ == '__main__':
    test_separate_common_items()
    test_reducemap()
    test_itermap()
    test_remove_duplicates()
    test_detect_duplicates()
    test_remove_common_prefix()
    test_all_equal()



from artemis.experiments.experiments import Experiment


def experiment_function(f):
    """
    Use this decorator (@experiment_function) on a function that you want to run.  e.g.

    .. code-block:: python

        @experiment_function
        def demo_my_experiment(a=1, b=2, c=3):
            ...

    This turns your function demo_my_experiment into an experiment.  It can still be called as a normal function, but
    it now has can also be called with the methods of an Experiment object (eg. demo_my_experiment.run()).

    """
    return ExperimentFunction()(f)


def experiment_root(f):
    """
    Use this decorator on a function that you want to build variants off of:

    .. code-block:: python

        @experiment_root
        def demo_my_experiment(a, b=2, c=3):
            ...

    The root experiment is not runnable by itself, and will not appear in the list in the browse experiments UI, but
    you can call ``demo_my_experiment.add_variant(...)`` to create runnable variants.
    """
    return ExperimentFunction(is_root=True)(f)


class ExperimentFunction(object):
    """
    This is the most general decorator.  You can use this to add details on the experiment.
    """

    def __init__(self, display_function=None, comparison_function=None, one_liner_function=None, is_root=False):
        """
        :param display_function: A function that takes the results (whatever your experiment returns) and displays them.
        :param comparison_function: A function that takes an OrderedDict<experiment_name, experiment_return_value>.
            You can optionally define this function to compare the results of different experiments.
            You can use call this via the UI with the compare_experiment_results command.
        :param one_liner_function: A function that takes your results and returns a 1 line string summarizing them.
        :param is_root: True to make this a root experiment - so that it is not listed to be run itself.
        """
        self.display_function = display_function
        self.comparison_function = comparison_function
        self.is_root = is_root
        self.one_liner_function = one_liner_function

    def __call__(self, f):
        f.is_base_experiment = True
        ex = Experiment(
            name=f.__name__,
            function=f,
            display_function=self.display_function,
            comparison_function = self.comparison_function,
            one_liner_function=self.one_liner_function,
            is_root=self.is_root
        )
        return ex